import csv
import logging
from enum import Enum
from typing import List

from brain_brew.representation.generic.source_file import SourceFile
from brain_brew.utils import list_of_str_to_lowercase, sort_dict


class CsvKeys(Enum):
    GUID = "guid"
    TAGS = "tags"


class CsvFile(SourceFile):
    file_location: str = ""
    _data: List[dict] = []
    column_headers: list = []

    def __init__(self, file):
        self.file_location = file
        self.read_file()

    @classmethod
    def from_file_loc(cls, file_loc) -> 'CsvFile':
        return cls(file_loc)

    def read_file(self):
        self._data = []

        with open(self.file_location, mode='r') as csv_file:
            csv_reader = csv.DictReader(csv_file)

            self.column_headers = list_of_str_to_lowercase(csv_reader.fieldnames)

            for row in csv_reader:
                self._data.append({key.lower(): row[key] for key in row})

    def write_file(self):
        logging.info(f"Writing to Csv '{self.file_location}'")
        with open(self.file_location, mode='w+') as csv_file:
            csv_writer = csv.DictWriter(csv_file, fieldnames=self.column_headers)

            csv_writer.writeheader()

            for row in self._data:
                csv_writer.writerow(row)

    def set_data(self, data_override):
        self._data = data_override
        self.column_headers = list(data_override[0].keys()) if data_override else []

    def set_data_from_superset(self, superset: List[dict], column_header_override=None):
        if column_header_override:
            self.column_headers = column_header_override

        data_to_set: List[dict] = []
        for row in superset:
            if not all(column in row for column in self.column_headers):
                continue
            new_row = {}
            for column in self.column_headers:
                new_row[column] = row[column]
            data_to_set.append(new_row)

        self.set_data(data_to_set)

    def get_data(self, deep_copy=False) -> List[dict]:
        return self.get_deep_copy(self._data) if deep_copy else self._data

    @staticmethod
    def to_filename_csv(filename: str) -> str:
        return filename + ".csv" if not filename.endswith(".csv") else filename

    @classmethod
    def formatted_file_location(cls, location):
        return cls.to_filename_csv(location)

    def sort_data(self, sort_by_keys, reverse_sort):
        self._data = sort_dict(self._data, sort_by_keys, reverse_sort)



from dataclasses import dataclass, field
from typing import Union, Optional

from brain_brew.configuration.build_config.build_task import BuildPartTask
from brain_brew.configuration.part_holder import PartHolder
from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.representation.json.crowd_anki_export import CrowdAnkiExport
from brain_brew.representation.json.wrappers_for_crowd_anki import CA_NOTE_MODELS, CA_NOTES, CA_MEDIA_FILES, \
    CA_CHILDREN, CA_TYPE
from brain_brew.representation.json.wrappers_for_crowd_anki import CrowdAnkiJsonWrapper
from brain_brew.representation.yaml.headers import Headers

headers_skip_keys = [CA_NOTE_MODELS, CA_NOTES, CA_MEDIA_FILES]
headers_default_values = {
    CA_TYPE: "Deck",
    CA_CHILDREN: [],
}


@dataclass
class HeadersFromCrowdAnki(BuildPartTask):
    @classmethod
    def task_name(cls) -> str:
        return r'headers_from_crowd_anki'

    @classmethod
    def task_regex(cls) -> str:
        return r'header[s]?_from_crowd_anki'

    @classmethod
    def yamale_schema(cls) -> str:
        return f'''\
            source: str()
            part_id: str()
            save_to_file: str(required=False)
        '''

    @dataclass
    class Representation(RepresentationBase):
        source: str
        part_id: str
        save_to_file: Optional[str] = field(default=None)

    @classmethod
    def from_repr(cls, data: Union[Representation, dict]):
        rep: cls.Representation = data if isinstance(data, cls.Representation) else cls.Representation.from_dict(data)
        return cls(
            ca_export=CrowdAnkiExport.create_or_get(rep.source),
            part_id=rep.part_id,
            save_to_file=rep.save_to_file
        )

    ca_export: CrowdAnkiExport
    part_id: str
    save_to_file: Optional[str]

    def execute(self):
        ca_wrapper: CrowdAnkiJsonWrapper = self.ca_export.json_data

        headers = Headers(self.crowd_anki_to_headers(ca_wrapper.data))

        PartHolder.override_or_create(self.part_id, self.save_to_file, headers)

    @staticmethod
    def crowd_anki_to_headers(ca_data: dict):
        return {key: value for key, value in ca_data.items()
                if key not in headers_skip_keys and key not in headers_default_values.keys()}



# def setup_csv_fm_config(csv: str, sort_by_columns: List[str] = None, reverse_sort: bool = None,
#                         note_model_name: str = None, derivatives: List[dict] = None):
#     cfm: dict = {
#         FILE: csv
#     }
#     if sort_by_columns is not None:
#         cfm.setdefault(SORT_BY_COLUMNS, sort_by_columns)
#     if reverse_sort is not None:
#         cfm.setdefault(REVERSE_SORT, reverse_sort)
#     if note_model_name is not None:
#         cfm.setdefault(NOTE_MODEL, note_model_name)
#     if derivatives is not None:
#         cfm.setdefault(DERIVATIVES, derivatives)
#
#     return cfm
#

# class TestConstructor:
#     @pytest.mark.parametrize("read_file_now, note_model_name, csv, sort_by_columns, reverse_sort", [
#         (False, "note_model.json", "first.csv", ["guid"], False),
#         (True, "model_model.json", "second.csv", ["guid", "note_model_name"], True),
#         (False, "note_model-json", "first.csv", ["guid"], False,)
#     ])
#     def test_runs_without_derivatives(self, read_file_now, note_model_name, csv,
#                                       sort_by_columns, reverse_sort):
#         get_new_file_manager()
#         config = setup_csv_fm_config(csv, sort_by_columns, reverse_sort, note_model_name=note_model_name)
#
#         def assert_csv(passed_file, read_now):
#             assert passed_file == csv
#             assert read_now == read_file_now
#
#         with patch.object(FileMappingDerivative, "create_derivative", return_value=None) as mock_derivatives, \
#                 patch.object(CsvFile, "create", side_effect=assert_csv) as mock_csv:
#
#             csv_fm = FileMapping(config, read_now=read_file_now)
#
#             assert isinstance(csv_fm, FileMapping)
#             assert csv_fm.reverse_sort == reverse_sort
#             assert csv_fm.sort_by_columns == sort_by_columns
#             assert csv_fm.note_model_name == note_model_name
#
#             assert mock_csv.call_count == 1
#             assert mock_derivatives.call_count == 0
#
#     @pytest.mark.parametrize("derivatives", [
#         [setup_csv_fm_config("test_csv.csv")],
#         [setup_csv_fm_config("test_csv.csv"), setup_csv_fm_config("second.csv")],
#         [setup_csv_fm_config("a.csv"), setup_csv_fm_config("b.csv"), setup_csv_fm_config("c.csv")],
#         [setup_csv_fm_config("a.csv", sort_by_columns=["word", "guid"], reverse_sort=True, note_model_name="d")],
#         [setup_csv_fm_config("test_csv.csv", derivatives=[setup_csv_fm_config("der_der.csv")])],
#     ])
#     def test_runs_with_derivatives(self, derivatives: list):
#         get_new_file_manager()
#         config = setup_csv_fm_config("test", [], False, note_model_name="nm", derivatives=derivatives.copy())
#         expected_call_count = len(derivatives)
#
#         def assert_der(passed_file, read_now):
#             der = derivatives.pop(0)
#             assert passed_file == der
#             assert read_now is False
#
#         with patch.object(FileMappingDerivative, "create_derivative", side_effect=assert_der) as mock_derivatives, \
#                 patch.object(CsvFile, "create", return_value=None):
#
#             csv_fm = FileMapping(config, read_now=False)
#
#             assert mock_derivatives.call_count == len(csv_fm.derivatives) == expected_call_count


# def csv_fixture_gen(csv_fix):
#     with patch.object(CsvFile, "create_or_get", return_value=csv_fix):
#         csv = FileMapping(**setup_csv_fm_config("", note_model_name="Test Model"))
#         csv.compile_data()
#         return csv
#
#
# @pytest.fixture()
# def csv_file_mapping1(csv_test1):
#     return csv_fixture_gen(csv_test1)
#
#
# @pytest.fixture()
# def csv_file_mapping2(csv_test2):
#     return csv_fixture_gen(csv_test2)
#
#
# @pytest.fixture()
# def csv_file_mapping3(csv_test3):
#     return csv_fixture_gen(csv_test3)
#
#
# @pytest.fixture()
# def csv_file_mapping_split1(csv_test1_split1):
#     return csv_fixture_gen(csv_test1_split1)
#
#
# @pytest.fixture()
# def csv_file_mapping_split1(csv_test1_split2):
#     return csv_fixture_gen(csv_test1_split2)
#
#
# @pytest.fixture()
# def csv_file_mapping2_missing_guids(csv_test2_missing_guids):
#     return csv_fixture_gen(csv_test2_missing_guids)
#
#
# class TestSetRelevantData:
#     def test_no_change(self, csv_file_mapping1: FileMapping, csv_file_mapping_split1: FileMapping):
#         assert csv_file_mapping1.data_set_has_changed is False
#
#         previous_data = csv_file_mapping1.compiled_data.copy()
#         csv_file_mapping1.set_relevant_data(csv_file_mapping_split1.compiled_data)
#
#         assert previous_data == csv_file_mapping1.compiled_data
#         assert csv_file_mapping1.data_set_has_changed is False
#
#     def test_change_but_no_extra(self, csv_file_mapping1: FileMapping, csv_file_mapping2: FileMapping):
#         assert csv_file_mapping1.data_set_has_changed is False
#         assert len(csv_file_mapping1.compiled_data) == 15
#
#         previous_data = copy.deepcopy(csv_file_mapping1.compiled_data)
#         csv_file_mapping1.set_relevant_data(csv_file_mapping2.compiled_data)
#
#         assert previous_data != csv_file_mapping1.compiled_data
#         assert csv_file_mapping1.data_set_has_changed is True
#         assert len(csv_file_mapping1.compiled_data) == 15
#
#     def test_change_extra_row(self, csv_file_mapping1: FileMapping, csv_file_mapping3: FileMapping):
#         assert csv_file_mapping1.data_set_has_changed is False
#         assert len(csv_file_mapping1.compiled_data) == 15
#
#         previous_data = copy.deepcopy(csv_file_mapping1.compiled_data.copy())
#         csv_file_mapping1.set_relevant_data(csv_file_mapping3.compiled_data)
#
#         assert previous_data != csv_file_mapping1.compiled_data
#         assert csv_file_mapping1.data_set_has_changed is True
#         assert len(csv_file_mapping1.compiled_data) == 16
#
#
# class TestCompileData:
#     num = 0
#
#     def get_num(self):
#         self.num += 1
#         return self.num
#
#     def test_when_missing_guids(self, csv_file_mapping2_missing_guids: FileMapping):
#         with patch("brain_brew.representation.configuration.csv_file_mapping.generate_anki_guid", wraps=self.get_num) as mock_guid:
#
#             csv_file_mapping2_missing_guids.compile_data()
#
#             assert csv_file_mapping2_missing_guids.data_set_has_changed is True
#             assert mock_guid.call_count == 9
#             assert list(csv_file_mapping2_missing_guids.compiled_data.keys()) == list(range(1, 10))

#  Tests still to do:
#
#  Top level needs a NoteModel, others do not




from collections import OrderedDict
from dataclasses import dataclass, field
from typing import List, Union, Dict, Set

from brain_brew.configuration.anki_field import AnkiField
from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.interfaces.media_container import MediaContainer
from brain_brew.representation.yaml.note_model_field import Field
from brain_brew.representation.yaml.note_model_template import Template
from brain_brew.representation.yaml.yaml_object import YamlObject
from brain_brew.utils import list_of_str_to_lowercase

# CrowdAnki
CROWDANKI_ID = AnkiField("crowdanki_uuid", "id")
CROWDANKI_TYPE = AnkiField("__type__", default_value="NoteModel")

# Shared
NAME = AnkiField("name")
ORDINAL = AnkiField("ord", "ordinal")

# Note Model
CSS = AnkiField("css")
LATEX_PRE = AnkiField("latexPre", "latex_pre",
                      default_value="\\documentclass[12pt]{article}\n\\special{papersize=3in,5in}\n\\usepackage{"
                                    "amssymb,amsmath}\n\\pagestyle{empty}\n\\setlength{\\parindent}{0in}\n\\begin{"
                                    "document}\n")
LATEX_POST = AnkiField("latexPost", "latex_post", default_value="\\end{document}")
REQUIRED_FIELDS_PER_TEMPLATE = AnkiField("req", "required_fields_per_template")
FIELDS = AnkiField("flds", "fields")
TEMPLATES = AnkiField("tmpls", "templates")
TAGS = AnkiField("tags", default_value=[])
SORT_FIELD_NUM = AnkiField("sortf", "sort_field_num", default_value=0)
IS_CLOZE = AnkiField("type", "is_cloze", default_value=False)
VERSION = AnkiField("vers", "version", default_value=[])

# Field
FONT = AnkiField("font", default_value="Liberation Sans")
MEDIA = AnkiField("media", default_value=[])
IS_RIGHT_TO_LEFT = AnkiField("rtl", "is_right_to_left", default_value=False)
FONT_SIZE = AnkiField("size", "font_size", default_value=20)
IS_STICKY = AnkiField("sticky", "is_sticky", default_value=False)

# Template
QUESTION_FORMAT = AnkiField("qfmt", "question_format")
ANSWER_FORMAT = AnkiField("afmt", "answer_format")
BROWSER_ANSWER_FORMAT = AnkiField("bafmt", "browser_answer_format", default_value="")
BROWSER_QUESTION_FORMAT = AnkiField("bqfmt", "browser_question_format", default_value="")
DECK_OVERRIDE_ID = AnkiField("did", "deck_override_id", default_value=None)


@dataclass
class NoteModel(YamlObject, MediaContainer, RepresentationBase):
    @dataclass
    class CrowdAnki(RepresentationBase):
        name: str
        crowdanki_uuid: str
        css: str
        req: List[list]
        flds: List[dict]
        tmpls: List[dict]
        latexPre: str = field(default=LATEX_PRE.default_value)
        latexPost: str = field(default=LATEX_POST.default_value)
        __type__: str = field(default=CROWDANKI_TYPE.default_value)
        tags: List[str] = field(default_factory=lambda: TAGS.default_value)
        sortf: int = field(default=SORT_FIELD_NUM.default_value)
        type: int = field(default=0)  # Is_Cloze Manually set to 0
        vers: list = field(default_factory=lambda: VERSION.default_value)

    name: str
    id: str
    css: str
    required_fields_per_template: List[list]
    fields: List[Field]
    templates: List[Template]

    latex_post: str = field(default=LATEX_POST.default_value)
    latex_pre: str = field(default=LATEX_PRE.default_value)
    sort_field_num: int = field(default=SORT_FIELD_NUM.default_value)
    is_cloze: bool = field(default=IS_CLOZE.default_value)
    crowdanki_type: str = field(default=CROWDANKI_TYPE.default_value)  # Should always be "NoteModel"
    tags: List[str] = field(default_factory=lambda: TAGS.default_value)  # Tags of the last added note
    version: list = field(default_factory=lambda: VERSION.default_value)  # Legacy version number. Deprecated in Anki

    @classmethod
    def from_yaml_file(cls, filename: str):
        data = cls.read_to_dict(filename)
        return cls(
            fields=[Field.from_dict(f) for f in data.pop(FIELDS.name)],
            templates=[Template.from_dict(t) for t in data.pop(TEMPLATES.name)],
            **data
        )

    @classmethod
    def from_crowdanki(cls, data: Union[CrowdAnki, dict]):  # TODO: field_whitelist, note_model_whitelist
        ca: cls.CrowdAnki = data if isinstance(data, cls.CrowdAnki) else cls.CrowdAnki.from_dict(data)
        return cls(
            fields=[Field.from_crowd_anki(f) for f in ca.flds],
            templates=[Template.from_crowdanki(t) for t in ca.tmpls],
            is_cloze=bool(ca.type),
            name=ca.name, css=ca.css, latex_pre=ca.latexPre, latex_post=ca.latexPost,
            required_fields_per_template=ca.req, tags=ca.tags, sort_field_num=ca.sortf, version=ca.vers,
            id=ca.crowdanki_uuid, crowdanki_type=ca.__type__
        )

    def encode_as_crowdanki(self) -> dict:
        data_dict = {
            NAME.anki_name: self.name,
            CROWDANKI_ID.anki_name: self.id,
            CSS.anki_name: self.css,
            REQUIRED_FIELDS_PER_TEMPLATE.anki_name: self.required_fields_per_template,
            LATEX_PRE.anki_name: self.latex_pre,
            LATEX_POST.anki_name: self.latex_post,
            SORT_FIELD_NUM.anki_name: self.sort_field_num,
            CROWDANKI_TYPE.anki_name: self.crowdanki_type,
            TAGS.anki_name: self.tags,
            VERSION.anki_name: self.version,
            IS_CLOZE.anki_name: 1 if self.is_cloze else 0
        }

        data_dict.setdefault(FIELDS.anki_name, [f.encode_as_crowdanki(num) for num, f in enumerate(self.fields)])
        data_dict.setdefault(TEMPLATES.anki_name, [t.encode_as_crowdanki(num) for num, t in enumerate(self.templates)])

        return OrderedDict(sorted(data_dict.items()))

    def encode(self) -> dict:
        data_dict: Dict[str, Union[str, list]] = {
            NAME.name: self.name,
            CROWDANKI_ID.name: self.id,
            CSS.name: self.css
        }

        SORT_FIELD_NUM.append_name_if_differs(data_dict, self.sort_field_num)
        IS_CLOZE.append_name_if_differs(data_dict, self.is_cloze)
        LATEX_PRE.append_name_if_differs(data_dict, self.latex_pre)
        LATEX_POST.append_name_if_differs(data_dict, self.latex_post)

        data_dict.setdefault(FIELDS.name, [f.encode_as_part() for f in self.fields])
        data_dict.setdefault(TEMPLATES.name, [t.encode() for t in self.templates])

        # Useless
        TAGS.append_name_if_differs(data_dict, self.tags)
        VERSION.append_name_if_differs(data_dict, self.version)
        CROWDANKI_TYPE.append_name_if_differs(data_dict, self.crowdanki_type)
        data_dict.setdefault(REQUIRED_FIELDS_PER_TEMPLATE.name, self.required_fields_per_template)

        return data_dict

    def get_all_media_references(self) -> Set[str]:
        all_media = set()
        for template in self.templates:
            all_media = all_media.union(template.get_all_media_references())

        return all_media

    @property
    def field_names_lowercase(self):
        return list_of_str_to_lowercase([f.name for f in self.fields])

    def check_field_overlap(self, fields_to_check: List[str]):
        fields_to_check = list_of_str_to_lowercase(fields_to_check)

        missing = [f for f in self.field_names_lowercase if f not in fields_to_check]

        return missing

    def check_field_extra(self, fields_to_check: List[str]):
        fields_to_check = list_of_str_to_lowercase(fields_to_check)

        return [f for f in fields_to_check if f not in self.field_names_lowercase]

    def zip_field_to_data(self, data: List[str]) -> dict:
        if len(self.fields) != len(data):
            raise Exception(
                f"Data of length {len(data)} cannot map to fields of length {len(self.field_names_lowercase)}")
        return dict(zip(self.field_names_lowercase, data))





from brain_brew.configuration.part_holder import PartHolder


def debug_write_part_to_file(part, filepath: str):
    dp = PartHolder("Blah", filepath, part)
    dp.save_to_file = filepath
    dp.write_to_file()



import logging
import os

import yamale
from yamale import YamaleError
from yamale.schema import Schema
from yamale.validators import DefaultValidators

validators = DefaultValidators.copy()


class YamlVerifier:
    __instance = None
    recipe_schema: Schema

    def __init__(self):
        if YamlVerifier.__instance is None:
            YamlVerifier.__instance = self
        else:
            raise Exception("Multiple YamlVerifiers created")

        path = os.path.join(os.path.dirname(__file__), "../schemas/recipe.yaml")
        self.recipe_schema = yamale.make_schema(path, parser='ruamel', validators=validators)

    @staticmethod
    def get_instance() -> 'YamlVerifier':
        return YamlVerifier.__instance

    def verify_recipe(self, filename):
        data = yamale.make_data(filename)
        try:
            yamale.validate(self.recipe_schema, data)
        except YamaleError as e:
            print('Validation failed!\n')
            for result in e.results:
                print("Error validating data '%s' with '%s'\n\t" % (result.data, result.schema))
                for error in result.errors:
                    print('\t%s' % error)
            exit(1)
        logging.info(f"Builder file {filename} is ✔ good")



from dataclasses import dataclass
from typing import Set, Dict, List, Tuple

from brain_brew.representation.generic.media_file import MediaFile
from brain_brew.representation.yaml.yaml_object import YamlObject
from brain_brew.utils import find_all_files_in_directory


@dataclass
class MediaGroup(YamlObject):
    media_files: Dict[str, MediaFile]

    def encode(self) -> list:
        return list(m.file_path for m in self.media_files.values())  # TODO: Use relative path for directory?

    @classmethod
    def from_yaml_file(cls, filename: str) -> 'MediaGroup':
        return cls(media_files=cls.from_full_path_list(cls.read_to_dict(filename)))

    @classmethod
    def from_directory(cls, directory: str, recursive: bool) -> 'MediaGroup':
        return cls(media_files=cls.from_full_path_list(find_all_files_in_directory(directory, recursive=recursive)))

    @classmethod
    def from_many(cls, groups: List['MediaGroup']) -> 'MediaGroup':
        files = list(set(file.file_path for group in groups for file in group.media_files.values()))
        return cls(media_files=cls.from_full_path_list(files))

    @staticmethod
    def from_full_path_list(known_files: list):
        files: Dict[str, MediaFile] = dict()

        for full_path in known_files:
            file = MediaFile.create_or_get(full_path)
            if file.filename not in files.keys():
                files[file.filename] = file
            else:
                raise NameError(f"Duplicate files with same filename '{file.filename}' in group")

        return files

    def remove_by_filename(self, filename: str):
        self.media_files.pop(filename, None)

    def filter_by_filenames(self, filenames: List[str], should_match: bool):
        for media_filename in self.media_files.keys():
            is_match = media_filename in filenames
            if is_match != should_match:
                self.remove_by_filename(media_filename)
        # TODO: Find all missing files

    def compare(self, other: 'MediaGroup') -> Tuple[Set[str], Set[str], Set[str]]:

        self_set = set(self.media_files)
        other_set = set(other.media_files)

        return self_set.intersection(other_set), self_set - other_set, other_set - self_set



class AnkiField:
    name: str
    anki_name: str
    default_value: any

    def __init__(self, anki_name, name=None, default_value=None):
        self.anki_name = anki_name
        self.name = name if name is not None else anki_name
        self.default_value = default_value

    def append_name_if_differs(self, dict_to_add_to: dict, value):
        if value != self.default_value:
            dict_to_add_to.setdefault(self.name, value)



import pytest

from brain_brew.representation.generic.csv_file import CsvFile
from tests.test_files import TestFiles


@pytest.fixture()
def csv_test1():
    return CsvFile(TestFiles.CsvFiles.TEST1)


@pytest.fixture()
def csv_test1_split1():
    return CsvFile(TestFiles.CsvFiles.TEST1_SPLIT1)


@pytest.fixture()
def csv_test1_split2():
    return CsvFile(TestFiles.CsvFiles.TEST1_SPLIT2)


@pytest.fixture()
def csv_test2():
    return CsvFile(TestFiles.CsvFiles.TEST2)


@pytest.fixture()
def csv_test3():
    return CsvFile(TestFiles.CsvFiles.TEST3)


@pytest.fixture()
def csv_test2_missing_guids():
    return CsvFile(TestFiles.CsvFiles.TEST2_MISSING_GUIDS)


@pytest.fixture()
def temp_csv_test1(tmpdir, csv_test1) -> CsvFile:
    file = tmpdir.mkdir("json").join("file.csv")
    file.write("blank")

    return CsvFile.create_or_get(file.strpath)


# class TestConstructor:
#     def test_runs(self, csv_test1):
#         assert isinstance(csv_test1, CsvFile)
#         assert csv_test1.file_location == TestFiles.CsvFiles.TEST1
#         assert "guid" in csv_test1.column_headers
#
#
# def test_to_filename_csv():
#     expected = "read-this-file.csv"
#
#     assert expected == CsvFile.to_filename_csv("read this file")
#     assert expected == CsvFile.to_filename_csv("read-this-file")
#     assert expected == CsvFile.to_filename_csv("read-this-file.csv")
#     assert expected == CsvFile.to_filename_csv("read          this        file")
#
#
# class TestWriteFile:
#     def test_runs(self, temp_csv_test1: CsvFile, csv_test1: CsvFile):
#         temp_csv_test1.write_file()
#         temp_csv_test1.read_file()
#
#         assert temp_csv_test1.get_data() == csv_test1.get_data()
#

# class TestSortData:
#     @pytest.mark.parametrize("columns, reverse, result_column, expected_results", [
#         (["guid"], False, "guid", [(0, "AAAA"), (1, "BBBB"), (2, "CCCC"), (14, "OOOO")]),
#         (["guid"], True, "guid", [(14, "AAAA"), (13, "BBBB"), (12, "CCCC"), (0, "OOOO")]),
#         (["english"], False, "english", [(0, "banana"), (1, "bird"), (2, "cat"), (14, "you")]),
#         (["english"], True, "english", [(14, "banana"), (13, "bird"), (12, "cat"), (0, "you")]),
#         (["tags"], False, "tags", [(0, "besttag"), (1, "funny"), (2, "tag2 tag3"), (13, ""), (14, "")]),
#         (["esperanto", "english"], False, "esperanto", [(0, "banano"), (1, "birdo"), (6, "vi"), (14, "")]),
#         (["esperanto", "guid"], False, "guid", [(7, "BBBB"), (14, "LLLL")]),
#     ])
#     def test_sort(self, csv_test1: CsvFile, columns, reverse, result_column, expected_results):
#         csv_test1.sort_data(columns, reverse)
#
#         sorted_data = csv_test1.get_data()
#
#         for result in expected_results:
#             assert sorted_data[result[0]][result_column] == result[1]
#
#     def test_insensitive(self):
#         pass



import logging
from typing import List, Set

from brain_brew.representation.generic.media_file import MediaFile
from brain_brew.representation.yaml.media_group import MediaGroup
from brain_brew.utils import create_path_if_not_exists


def save_media_groups_to_location(
        parts: List[MediaGroup],
        folder: str,
        clear_folder: bool,
        recursive: bool
) -> Set[MediaFile]:

    existing_media_group = MediaGroup.from_directory(folder, recursive)
    all_media_group = MediaGroup.from_many(parts)

    in_both, to_move, to_delete = all_media_group.compare(existing_media_group)

    create_path_if_not_exists(folder)
    for filename, media_file in all_media_group.media_files.items():
        if filename in in_both:
            media_file.copy_self_to_target(existing_media_group.media_files[filename].file_path)
            # TODO: Check if copying is needed?
        elif filename in to_move:
            media_file.copy_self_to_target(folder)

    if clear_folder and to_delete:
        deleted = '\n'.join(to_delete)
        logging.warning(f"Deleting extra files in media folder '{folder}':\n{'-'*20}\n{deleted}\n{'-'*20}")
        for delete_name in to_delete:
            existing_media_group.media_files[delete_name].delete_self()

    return set(all_media_group.media_files.values())



# from typing import List
# from unittest.mock import patch
#
# import pytest
#
# from brain_brew.build_tasks.source_csv import SourceCsv, SourceCsvKeys
# from brain_brew.constants.deckpart_keys import DeckPartNoteKeys
# from brain_brew.representation.configuration.csv_file_mapping import FileMapping
# from brain_brew.representation.configuration.note_model_mapping import NoteModelMapping
# from brain_brew.representation.generic.csv_file import CsvFile
# from brain_brew.representation.generic.source_file import SourceFile
# from brain_brew.representation.json.part_notes import DeckPartNotes
# from tests.representation.json.test_part_notes import dp_notes_test1
# from tests.representation.configuration.test_note_model_mapping import setup_nmm_config
# from tests.representation.configuration.test_csv_file_mapping import setup_csv_fm_config
#
#
# def setup_source_csv_config(notes: str, nmm: list, csv_mappings: list):
#     return {
#         SourceCsvKeys.NOTES.value: notes,
#         SourceCsvKeys.NOTE_MODEL_MAPPINGS.value: nmm,
#         SourceCsvKeys.CSV_MAPPINGS.value: csv_mappings
#     }
#
#
# def get_csv_default(notes: DeckPartNotes, nmm: List[NoteModelMapping], csv_maps: List[FileMapping]) -> SourceCsv:
#     csv_source = SourceCsv(setup_source_csv_config("", [], []), read_now=False)
#
#     csv_source.notes = notes
#     csv_source.note_model_mappings_dict = {nm_map.note_model.name: nm_map for nm_map in nmm}
#     csv_source.csv_file_mappings = csv_maps
#
#     return csv_source
#
#
# @pytest.fixture()
# def csv_source_test1(dp_notes_test1, nmm_test1, csv_file_mapping1) -> SourceCsv:
#     return get_csv_default(dp_notes_test1, [nmm_test1], [csv_file_mapping1])
#
#
# @pytest.fixture()
# def csv_source_test1_split1(csv_source_default, csv_test1_split1, dp_notes_test1) -> SourceCsv:
#     csv_source_default.csv_file = csv_test1_split1
#     csv_source_default.notes = dp_notes_test1
#     return csv_source_default
#
#
# @pytest.fixture()
# def csv_source_test1_split2(csv_source_default2, csv_test1_split2, dp_notes_test2) -> SourceCsv:
#     csv_source_default2.csv_file = csv_test1_split2
#     csv_source_default2.notes = dp_notes_test1
#     return csv_source_default2
#
#
# @pytest.fixture()
# def csv_source_test2(dp_notes_test2, nmm_test1, csv_file_mapping2) -> SourceCsv:
#     return get_csv_default(dp_notes_test2, [nmm_test1], [csv_file_mapping2])
#
#
# # @pytest.fixture()
# # def temp_csv_source(global_config, tmpdir) -> SourceCsv:
# #     file = tmpdir.mkdir("notes").join("file.csv")
# #     file.write("test,1,2,3")
#
#
# class TestConstructor:
#     def test_runs(self):
#         source_csv = get_csv_default(None, [], [])
#         assert isinstance(source_csv, SourceCsv)
#
#     @pytest.mark.parametrize("notes, model, columns, personal_fields, csv_file", [
#         ("notes.json", "Test Model", {"a": "b"}, ["extra"], "file.csv")
#     ])
#     def test_calls_correctly(self, notes, model, columns, personal_fields, csv_file, nmm_test1):
#         nmm_config = [setup_nmm_config(model, columns, personal_fields)]
#         csv_config = [setup_csv_fm_config(csv_file, note_model_name=model)]
#
#         def assert_csv(config, read_now):
#             assert config in csv_config
#             assert read_now is False
#
#         def assert_nmm(config, read_now):
#             assert config in nmm_config
#             assert read_now is False
#
#         def assert_dpn(config, read_now):
#             assert config == notes
#             assert read_now is False
#
#         with patch.object(FileMapping, "__init__", side_effect=assert_csv), \
#                 patch.object(NoteModelMapping, "__init__", side_effect=assert_nmm), \
#                 patch.object(NoteModelMapping, "note_model"), \
#                 patch.object(DeckPartNotes, "create", side_effect=assert_dpn):
#
#             #nmm_mock.return_value = False
#
#             source_csv = SourceCsv(setup_source_csv_config(
#                 notes,
#                 nmm_config,
#                 csv_config
#             ), read_now=False)
#
#
#     # def test_missing_non_required_columns
#
#
# class TestSourceToDeckParts:
#     def test_runs_first(self, csv_source_test1, dp_notes_test1, csv_source_test2, dp_notes_test2):
#         self.run_s2dp(csv_source_test1, dp_notes_test1)
#         self.run_s2dp(csv_source_test2, dp_notes_test2)
#
#     @staticmethod
#     def run_s2dp(csv_source: SourceCsv, dp_notes: DeckPartNotes):
#         def assert_format(notes_data):
#             assert notes_data == dp_notes.get_data()[DeckPartNoteKeys.NOTES.value]
#
#         with patch.object(DeckPartNotes, "set_data", side_effect=assert_format) as mock_set_data:
#             csv_source.source_to_parts()
#             assert mock_set_data.call_count == 1
#
#
# class TestDeckPartsToSource:
#     def test_runs_with_no_change(self, csv_source_test1, csv_test1, csv_source_test2, csv_test2):
#
#         self.run_dpts(csv_source_test1, csv_test1)
#         self.run_dpts(csv_source_test2, csv_test2)
#
#     @staticmethod
#     def run_dpts(csv_source: SourceCsv, csv_file: CsvFile):
#         def assert_format(source_data):
#             assert source_data == csv_file.get_data()
#
#         with patch.object(SourceFile, "set_data", side_effect=assert_format) as mock_set_data:
#             csv_source.parts_to_source()
#             assert csv_source.csv_file_mappings[0].data_set_has_changed is False
#
#             csv_source.csv_file_mappings[0].data_set_has_changed = True
#             csv_source.csv_file_mappings[0].write_file_on_close()
#             assert mock_set_data.call_count == 1
#



import sys
from argparse import ArgumentParser


class BBArgumentReader(ArgumentParser):
    def __init__(self):
        super().__init__(
            description='Manage Flashcards by Transforming them to various types'
        )

        self._set_parser_arguments()

    def _set_parser_arguments(self):
        self.add_argument(
            "recipe",
            metavar="recipe",
            type=str,
            help="Yaml file to use as the recipe"
        )
        self.add_argument(
            "--config", "--global-config", "-c",
            action="store",
            dest="config_file",
            default=None,
            type=str,
            help="Global config file to use"
        )
        self.add_argument(
            "--verify", "-v",
            action="store_true",
            dest="verify_only",
            default=False,
            help="Only verify the recipe contents, without running it."
        )

    def get_parsed(self, override_args=None):
        parsed_args = self.parse_args(args=override_args)

        # Required
        recipe = self.error_if_blank(parsed_args.recipe)

        # Optional
        config_file = parsed_args.config_file
        verify_only = parsed_args.verify_only

        return recipe, config_file, verify_only

    def error_if_blank(self, arg):
        if arg == "" or arg is None:
            self.error("Required argument missing")
        return arg

    def error(self, message):
        sys.stderr.write('error: %s\n' % message)
        self.print_help()
        sys.exit(2)





class TestFiles:
    class Headers:
        LOC = "tests/test_files/deck_parts/headers/"

        FIRST = "default header"
        FIRST_COMPLETE = "default-header.json"

    class NoteFiles:
        LOC = "tests/test_files/deck_parts/"

        TEST1_NO_GROUPING_OR_SHARED_TAGS = "csvtonotes1_withnogroupingorsharedtags.json"
        TEST1_WITH_GROUPING = "csvtonotes1_withgrouping.json"
        TEST1_WITH_SHARED_TAGS = "csvtonotes1_withsharedtags.json"
        TEST1_WITH_SHARED_TAGS_EMPTY_AND_GROUPING = "csvtonotes1_withsharedtagsandgrouping_butnothingtogroup.json"
        TEST2_WITH_SHARED_TAGS_AND_GROUPING = "csvtonotes2_withsharedtagsandgrouping.json"

    class CrowdAnkiNoteModels:
        LOC = "tests/test_files/deck_parts/note_models/"

        TEST = "Test Model"
        TEST_COMPLETE = LOC + "Test-Model.json"

        LL_WORD = "LL Word"
        LL_WORD_COMPLETE = LOC + "LL-Word.json"

        LL_WORD_COMPLETE_ONLY_REQUIRED = LOC + "LL-Word-Only-Required.json"

        LL_WORD_COMPLETE_NO_DEFAULTS = LOC + "LL-Word-No-Defaults.json"

    class NoteModels:
        LOC = "tests/test_files/deck_parts/yaml/note_models/"

        LL_WORD = LOC + "LL-Word.yaml"

        LL_WORD_ONLY_REQUIRED = LOC + "LL-Word-Only-Required.yaml"

        LL_WORD_NO_DEFAULTS = LOC + "LL-Word-No-Defaults.yaml"

    class CsvFiles:
        LOC = "tests/test_files/csv/"

        TEST1 = LOC + "test1.csv"
        TEST1_SPLIT1 = LOC + "test1_split1.csv"
        TEST1_SPLIT2 = LOC + "test1_split2.csv"
        TEST2 = LOC + "test2.csv"
        TEST2_MISSING_GUIDS = LOC + "test2_missing_guids.csv"
        TEST3 = LOC + "test3.csv"

    class CrowdAnkiExport:
        LOC = "tests/test_files/crowd_anki/"

        TEST1_FOLDER = LOC + "crowdanki_example_1/"
        TEST1_FOLDER_WITHOUT_SLASH = LOC + "crowdanki_example_1"
        TEST1_JSON = TEST1_FOLDER + "deck.json"

    class BuildConfig:
        LOC = "tests/test_files/build_files/"

        ONE_OF_EACH_TYPE = LOC + "builder1.yaml"

    class MediaFiles:
        LOC = "tests/test_files/media_files/"

    class YamlNotes:
        LOC = "tests/test_files/yaml/notes/"

        TEST1 = LOC + "note1.yaml"



from dataclasses import dataclass
from typing import Optional, Union

from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.interfaces.yamale_verifyable import YamlRepr
from brain_brew.representation.yaml.notes import Notes, Note


@dataclass
class NotesOverride(YamlRepr):
    @classmethod
    def task_name(cls) -> str:
        return r"notes_override"

    @classmethod
    def yamale_schema(cls) -> str:
        return f'''\
            note_model: str(required=False)
        '''

    @dataclass
    class Representation(RepresentationBase):
        note_model: Optional[str]

    @classmethod
    def from_repr(cls, data: Union[Representation, dict]):
        rep: cls.Representation = data if isinstance(data, cls.Representation) else cls.Representation.from_dict(data)
        return cls(
            note_model=rep.note_model
        )

    note_model: Optional[str]

    def override(self, note: Note):
        if self.note_model:
            note.note_model = self.note_model

        return note



import pytest

from brain_brew.utils import find_media_in_field, str_to_lowercase_no_separators, split_tags


class TestFindMedia:
    @pytest.mark.parametrize("field_value, expected_results", [
        (r'<img src="image.png">', ["image.png"]),
        (r'<img src="image.png" >', ["image.png"]),
        (r'< img src="image.png">', ["image.png"]),
        (r'<     img src="image.png">', ["image.png"]),
        (r'<img class="myamainzingclass" src="image.png" >', ["image.png"]),
        (r'<img class=noquotesclass src="image.png" >', ["image.png"]),
        (r'<img src="image.png" class=classattheendforsomereason>', ["image.png"]),
        (r'words in the field <img src="image.png" > end other stuff', ["image.png"]),
        (r'<img src="ug-map-saint_barthelemy.png" />', ["ug-map-saint_barthelemy.png"]),
        (r'<img src="ug-map-saint_barthelemy.png" /><img src="image.png">',
            ["ug-map-saint_barthelemy.png", "image.png"]),
        (r'[sound:test.mp3]', ["test.mp3"]),
        (r'[sound:test.mp3][sound:othersound.mp3]', ["test.mp3", "othersound.mp3"]),
        (r'[sound:test.mp3]                       [sound:othersound.mp3]', ["test.mp3", "othersound.mp3"]),
        (r'words in the field [sound:test.mp3] other stuff too [sound:othersound.mp3] end', ["test.mp3", "othersound.mp3"]),
        (r'[sound:unfinished-bracket.mp3', [])
    ])
    def test_find_media_in_field(self, field_value, expected_results):
        assert find_media_in_field(field_value) == expected_results


class TestHelperFunctions:
    @pytest.mark.parametrize("str_to_tidy", [
        'Generate Csv Blah Blah',
        'Generate__Csv_Blah-Blah',
        'Generate      Csv Blah Blah',
        'generateCsvBlahBlah'
    ])
    def test_remove_spacers_from_str(self, str_to_tidy):
        assert str_to_lowercase_no_separators(str_to_tidy) == "generatecsvblahblah"


class TestSplitTags:
    @pytest.mark.parametrize("str_to_split, expected_result", [
        ("tags1, tags2", ["tags1", "tags2"]),
        ("tags1 tags2", ["tags1", "tags2"]),
        ("tags1; tags2", ["tags1", "tags2"]),
        ("tags1      tags2", ["tags1", "tags2"]),
        ("tags1, tags2, tags3, tags4, tags5, tags6, tags7, tags8, tags9",
            ["tags1", "tags2", "tags3", "tags4", "tags5", "tags6", "tags7", "tags8", "tags9"]),
        ("tags1, tags2; tags3 tags4      tags5,     tags6;    tags7    tags8, tags9",
            ["tags1", "tags2", "tags3", "tags4", "tags5", "tags6", "tags7", "tags8", "tags9"]),
        ("tags1,tags2", ["tags1", "tags2"]),
        ("tags1;tags2", ["tags1", "tags2"]),
        ("tags1,    tags2", ["tags1", "tags2"]),
        ("tags1;    tags2", ["tags1", "tags2"]),
    ])
    def test_runs(self, str_to_split, expected_result):
        assert split_tags(str_to_split) == expected_result


# class TestJoinTags:
#     @pytest.mark.parametrize("join_with, expected_result", [
#         (", ", "test, test1, test2")
#     ])
#     def test_joins(self, global_config, join_with, expected_result):
#         list_to_join = ["test", "test1", "test2"]
#         global_config.flags.join_values_with = join_with
#
#         assert join_tags(list_to_join) == expected_result



import os
import shutil
from dataclasses import dataclass, field

from brain_brew.representation.generic.source_file import SourceFile
from brain_brew.utils import filename_from_full_path


@dataclass
class MediaFile(SourceFile):
    file_path: str
    filename: str = field(init=False)

    def __post_init__(self):
        self.filename = filename_from_full_path(self.file_path)

    @classmethod
    def from_file_loc(cls, file_loc) -> 'MediaFile':
        return cls(file_loc)

    def __repr__(self):
        return f"MediaFile({self.file_path})"

    def __hash__(self):
        return hash(self.__repr__())

    def copy_self_to_target(self, target: str):
        shutil.copy2(self.file_path, target)

    def delete_self(self):
        os.remove(self.file_path)



import logging

from brain_brew.configuration.argument_reader import BBArgumentReader
from brain_brew.configuration.build_config.top_level_builder import TopLevelBuilder
from brain_brew.configuration.file_manager import FileManager
from brain_brew.configuration.global_config import GlobalConfig
# sys.path.append(os.path.join(os.path.dirname(__file__), "dist"))
# sys.path.append(os.path.dirname(__file__))
from brain_brew.configuration.yaml_verifier import YamlVerifier


def main():
    logging.basicConfig(level=logging.DEBUG)

    # Read in Arguments
    argument_reader = BBArgumentReader()
    recipe_file_name, global_config_file, verify_only = argument_reader.get_parsed()

    # Read in Global Config File
    GlobalConfig.from_yaml_file(global_config_file) if global_config_file else GlobalConfig.from_yaml_file()
    FileManager()

    # Parse Build Config File
    YamlVerifier()
    recipe = TopLevelBuilder.parse_and_read(recipe_file_name, verify_only)

    if not verify_only:
        recipe.execute()


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from typing import Optional, Union

from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.interfaces.yamale_verifyable import YamlRepr
from brain_brew.representation.generic.html_file import HTMLFile
from brain_brew.representation.yaml.headers import Headers


@dataclass
class HeadersOverride(YamlRepr):
    @classmethod
    def task_name(cls) -> str:
        return r"headers_override"

    @classmethod
    def yamale_schema(cls) -> str:
        return f'''\
            deck_description_html_file: str(required=False)
        '''

    @dataclass
    class Representation(RepresentationBase):
        deck_description_html_file: Optional[str]

    @classmethod
    def from_repr(cls, data: Union[Representation, dict]):
        rep: cls.Representation = data if isinstance(data, cls.Representation) else cls.Representation.from_dict(data)
        return cls(
            deck_desc_html_file=HTMLFile.create_or_get(rep.deck_description_html_file)
        )

    deck_desc_html_file: Optional[HTMLFile]

    def override(self, header: Headers):
        if self.deck_desc_html_file:
            header.description = self.deck_desc_html_file.get_data(deep_copy=True)

        return header



from dataclasses import dataclass
from typing import Union

from brain_brew.build_tasks.deck_parts.from_yaml_part import FromYamlPartBase
from brain_brew.build_tasks.overrides.headers_override import HeadersOverride
from brain_brew.configuration.build_config.build_task import BuildPartTask
from brain_brew.configuration.part_holder import PartHolder
from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.representation.yaml.headers import Headers


@dataclass
class HeadersFromYamlPart(BuildPartTask):
    @classmethod
    def yamale_schema(cls) -> str:
        return f'''\
            part_id: str()
            file: str()
            override: include('{HeadersOverride.task_name()}', required=False)
        '''

    @classmethod
    def yamale_dependencies(cls) -> set:
        return {HeadersOverride}

    @classmethod
    def task_name(cls) -> str:
        return r'headers_from_yaml_part'

    @classmethod
    def task_regex(cls) -> str:
        return r'header[s]?_from_yaml_part'

    @dataclass
    class Representation(RepresentationBase):
        part_id: str
        file: str
        override: dict

    @classmethod
    def from_repr(cls, data: Union[Representation, dict]):
        rep: cls.Representation = data if isinstance(data, cls.Representation) else cls.Representation.from_dict(data)

        return cls(
            headers=PartHolder.override_or_create(
                part_id=rep.part_id,
                save_to_file=None,
                part=Headers.from_yaml_file(rep.file)
            ).part,
            override=HeadersOverride.from_repr(rep.override)
        )

    headers: Headers
    override: HeadersOverride

    def execute(self):
        if self.override:
            self.headers = self.override.override(self.headers)



from typing import Dict, Union

from brain_brew.configuration.part_holder import PartHolder
from brain_brew.representation.generic.source_file import SourceFile
from brain_brew.representation.yaml.yaml_object import YamlObject


class FileManager:
    __instance = None

    known_files_dict: Dict[str, SourceFile]
    known_parts: Dict[str, PartHolder[YamlObject]]

    def __init__(self):
        if FileManager.__instance is None:
            FileManager.__instance = self
        else:
            raise Exception("Multiple FileManagers created")

        self.known_files_dict = {}
        self.known_parts = {}

    @staticmethod
    def get_instance() -> 'FileManager':
        return FileManager.__instance

    @staticmethod
    def clear_instance():
        if FileManager.__instance:
            FileManager.__instance = None

    # Source Files

    def register_file(self, full_path, file):
        if full_path in self.known_files_dict:
            raise FileExistsError(f"File already known to FileManager, cannot be registered twice: {full_path}")
        self.known_files_dict.setdefault(full_path, file)

    def file_if_exists(self, file_location) -> Union[SourceFile, None]:
        if file_location in self.known_files_dict.keys():
            return self.known_files_dict[file_location]
        return None

    # Deck Parts

    def register_part(self, dp: PartHolder) -> PartHolder:
        if dp.part_id in self.known_parts:
            raise KeyError(f"Cannot use same name '{dp.part_id}' for multiple Deck Parts")
        self.known_parts.setdefault(dp.part_id, dp)
        return dp

    def get_part_if_exists(self, dp_name) -> Union[PartHolder[YamlObject], None]:
        return self.known_parts.get(dp_name)

    def get_part(self, name: str):
        if name not in self.known_parts:
            raise KeyError(f"Cannot find Deck Part '{name}'")
        return self.known_parts[name]



from argparse import ArgumentParser
from unittest.mock import patch

import pytest

from brain_brew.configuration.argument_reader import BBArgumentReader


@pytest.fixture()
def arg_reader_test1():
    return BBArgumentReader()


def test_constructor(arg_reader_test1):
    assert isinstance(arg_reader_test1, BBArgumentReader)
    assert isinstance(arg_reader_test1, ArgumentParser)


class TestArguments:
    @pytest.mark.parametrize("arguments", [
        (["test_recipe.yaml", "--config"]),
        (["test_recipe.yaml", "config_file.yaml", "--config"]),
        (["--config", "config_file.yaml"]),
        ([""]),
        ([])
    ])
    def test_broken_arguments(self, arg_reader_test1, arguments):
        def raise_exit(message):
            raise SystemExit

        with pytest.raises(SystemExit):
            with patch.object(BBArgumentReader, "error", side_effect=raise_exit):
                parsed_args = arg_reader_test1.get_parsed(arguments)

    @pytest.mark.parametrize("arguments, recipe, config_file, verify_only", [
        (["test_recipe.yaml"], "test_recipe.yaml", None, False),
        (["test_recipe.yaml", "--verify"], "test_recipe.yaml", None, True),
        (["test_recipe.yaml", "-v"], "test_recipe.yaml", None, True),
        (["test_recipe.yaml", "--config", "other_config.yaml"], "test_recipe.yaml", "other_config.yaml", False),
        (["test_recipe.yaml", "--config", "other_config.yaml", "-v"], "test_recipe.yaml", "other_config.yaml", True),
    ])
    def test_correct_arguments(self, arg_reader_test1, arguments, recipe, config_file, verify_only):
        parsed_args = arg_reader_test1.parse_args(arguments)

        assert parsed_args.recipe == recipe
        assert parsed_args.config_file == config_file
        assert parsed_args.verify_only == verify_only



import pytest

from brain_brew.representation.generic.media_file import MediaFile


@pytest.fixture()
def media_file_test1() -> MediaFile:
    return MediaFile("place/loc/file.txt")


class TestConstructor:
    def test_without_override(self):
        loc = "place/loc/file.txt"

        media_file = MediaFile(loc)

        assert isinstance(media_file, MediaFile)
        assert media_file.file_path == loc
        assert media_file.filename == "file.txt"


class TestCopy:
    def test_copies_file(self, tmpdir):
        source_dir = tmpdir.mkdir("source")
        source = source_dir.join("test.txt")
        source.write("test content")
        assert len(source_dir.listdir()) == 1

        target_dir = tmpdir.mkdir("target")
        target = target_dir.join("test.txt")
        assert len(target_dir.listdir()) == 0

        media_file = MediaFile(str(source))
        media_file.copy_self_to_target(str(target))

        assert len(target_dir.listdir()) == len(source_dir.listdir()) == 1
        assert target.read() == "test content"



from dataclasses import dataclass, field
from typing import List, Dict, Union

from brain_brew.build_tasks.csvs.shared_base_csvs import SharedBaseCsvs
from brain_brew.configuration.build_config.build_task import TopLevelBuildTask
from brain_brew.configuration.part_holder import PartHolder
from brain_brew.representation.yaml.notes import Notes, Note
from brain_brew.transformers.file_mapping import FileMapping
from brain_brew.transformers.note_model_mapping import NoteModelMapping
from brain_brew.utils import join_tags


@dataclass
class CsvsGenerate(SharedBaseCsvs, TopLevelBuildTask):
    @classmethod
    def task_name(cls) -> str:
        return r'generate_csvs'

    @classmethod
    def task_regex(cls) -> str:
        return r'generate_csv[s]?'

    @classmethod
    def yamale_schema(cls) -> str:  # TODO: Use NotesOverride here, just as in NotesToCrowdAnki
        return f'''\
            notes: str()
            note_model_mappings: list(include('{NoteModelMapping.task_name()}'))
            file_mappings: list(include('{FileMapping.task_name()}'))
        '''

    @classmethod
    def yamale_dependencies(cls) -> set:
        return {NoteModelMapping, FileMapping}

    notes: PartHolder[Notes]  # TODO: Accept Multiple Note Parts

    @dataclass
    class Representation(SharedBaseCsvs.Representation):
        notes: str

    @classmethod
    def from_repr(cls, data: Union[Representation, dict]):
        rep: cls.Representation = data if isinstance(data, cls.Representation) else cls.Representation.from_dict(data)
        return cls(
            notes=PartHolder.from_file_manager(rep.notes),
            file_mappings=rep.get_file_mappings(),
            note_model_mappings=dict(*map(cls.map_nmm, rep.note_model_mappings))
        )

    def execute(self):
        self.verify_contents()

        notes: List[Note] = self.notes.part.get_sorted_notes_copy(
            sort_by_keys=[], reverse_sort=False, case_insensitive_sort=False)
        self.verify_notes_match_note_model_mappings(notes)

        csv_data: List[dict] = [self.note_to_csv_row(note, self.note_model_mappings) for note in notes]
        rows_by_guid = {row["guid"]: row for row in csv_data}

        for fm in self.file_mappings:
            fm.compile_data()
            fm.set_relevant_data(rows_by_guid)
            fm.write_file_on_close()

    def verify_notes_match_note_model_mappings(self, notes: List[Note]):
        note_models_used = {note.note_model for note in notes}
        errors = [TypeError(f"Unknown note model type '{model}' in deck part '{self.notes.part_id}'. "
                            f"Add mapping for that model.")
                  for model in note_models_used if model not in self.note_model_mappings.keys()]

        if errors:
            raise Exception(errors)

    @staticmethod
    def note_to_csv_row(note: Note, note_model_mappings: Dict[str, NoteModelMapping]) -> dict:
        nm_name = note.note_model
        row = note_model_mappings[nm_name].note_models[nm_name].part.zip_field_to_data(note.fields)
        row["guid"] = note.guid
        row["tags"] = join_tags(note.tags)
        # TODO: Flags?

        return note_model_mappings[nm_name].note_fields_map_to_csv_row(row)



from dataclasses import dataclass, field
from typing import Optional, Union

from brain_brew.configuration.build_config.build_task import BuildPartTask
from brain_brew.configuration.part_holder import PartHolder
from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.representation.json.crowd_anki_export import CrowdAnkiExport
from brain_brew.representation.json.wrappers_for_crowd_anki import CrowdAnkiJsonWrapper
from brain_brew.representation.yaml.note_model import NoteModel


@dataclass
class NoteModelsFromCrowdAnki(BuildPartTask):
    @classmethod
    def task_name(cls) -> str:
        return r'note_models_from_crowd_anki'

    @classmethod
    def task_regex(cls) -> str:
        return r'note_model[s]?_from_crowd_anki'

    @classmethod
    def yamale_schema(cls) -> str:
        return f'''\
            source: str()
            part_id: str()
            model_name: str(required=False)
            save_to_file: str(required=False)
        '''

    @dataclass
    class Representation(RepresentationBase):
        source: str
        part_id: str
        model_name: Optional[str] = field(default=None)
        save_to_file: Optional[str] = field(default=None)
        # TODO: fields: Optional[List[str]]
        # TODO: templates: Optional[List[str]]

    @classmethod
    def from_repr(cls, data: Union[Representation, dict]):
        rep: cls.Representation = data if isinstance(data, cls.Representation) else cls.Representation.from_dict(data)
        return cls(
            ca_export=CrowdAnkiExport.create_or_get(rep.source),
            part_id=rep.part_id,
            model_name=rep.model_name or rep.part_id,
            save_to_file=rep.save_to_file
        )

    ca_export: CrowdAnkiExport
    model_name: str
    part_id: str
    save_to_file: Optional[str]

    def execute(self):
        ca_wrapper: CrowdAnkiJsonWrapper = self.ca_export.json_data

        note_models_dict = {model.get('name'): model for model in ca_wrapper.note_models}

        if self.model_name not in note_models_dict:
            raise ReferenceError(f"Missing Note Model '{self.model_name}' in CrowdAnki file")

        part = NoteModel.from_crowdanki(note_models_dict[self.model_name])
        PartHolder.override_or_create(self.part_id, self.save_to_file, part)



from dataclasses import dataclass
from enum import Enum
from typing import List, Union, Dict

from brain_brew.configuration.part_holder import PartHolder
from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.interfaces.yamale_verifyable import YamlRepr
from brain_brew.representation.yaml.note_model import NoteModel
from brain_brew.representation.yaml.notes import GUID, TAGS
from brain_brew.utils import single_item_to_list


class FieldMapping:
    class FieldMappingType(Enum):
        COLUMN = "column"
        PERSONAL_FIELD = "personal_field"
        DEFAULT = "default"

        @classmethod
        def values(cls):
            return set(it.value for it in cls)

    type: FieldMappingType
    value: str
    field_name: str

    def __init__(self, field_type: FieldMappingType, field_name: str, value: str):
        self.type = field_type
        self.field_name = field_name.lower()

        if self.type == FieldMapping.FieldMappingType.COLUMN:
            self.value = value.lower()
        else:
            self.value = value


@dataclass
class NoteModelMapping(YamlRepr):
    @classmethod
    def task_name(cls) -> str:
        return r'note_model_mapping'

    @classmethod
    def yamale_schema(cls) -> str:
        return f'''\
            note_models: any(list(str()), str())
            columns_to_fields: map(str(), key=str())
            personal_fields: list(str())
        '''

    @dataclass
    class Representation(RepresentationBase):
        note_models: Union[str, list]
        columns_to_fields: Dict[str, str]
        personal_fields: List[str]

    note_models: Dict[str, PartHolder[NoteModel]]
    columns: List[FieldMapping]
    personal_fields: List[FieldMapping]

    required_fields_definitions = [GUID, TAGS]

    @classmethod
    def from_repr(cls, data: Union[Representation, dict]):
        rep: cls.Representation = data if isinstance(data, cls.Representation) else cls.Representation.from_dict(data)
        note_models = [PartHolder.from_file_manager(model) for model in single_item_to_list(rep.note_models)]

        return cls(
            columns=[FieldMapping(
                field_type=FieldMapping.FieldMappingType.COLUMN,
                field_name=field,
                value=key) for key, field in rep.columns_to_fields.items()],
            personal_fields=[FieldMapping(
                field_type=FieldMapping.FieldMappingType.PERSONAL_FIELD,
                field_name=field,
                value="") for field in rep.personal_fields],
            note_models=dict(map(lambda nm: (nm.part_id, nm), note_models))
        )

    def get_note_model_mapping_dict(self):
        return {model: self for model in self.note_models}

    def verify_contents(self):
        errors = []

        extra_fields = [field.field_name for field in self.columns
                        if field.field_name not in self.required_fields_definitions]

        for holder in self.note_models.values():
            model: NoteModel = holder.part

            # Check for Required Fields
            missing = []
            for req in self.required_fields_definitions:
                if req not in [field.value for field in self.columns]:
                    missing.append(req)

            if missing:
                errors.append(KeyError(f"""Note model(s) "{holder.part_id}" to Csv config error: \
                                   Definitions for fields {missing} are required."""))

            # Check Fields Align with Note Type
            missing = model.check_field_overlap(
                [field.field_name for field in self.columns
                 if field.field_name not in self.required_fields_definitions]
            )
            missing = [m for m in missing if m not in [field.field_name for field in self.personal_fields]]

            if missing:
                errors.append(
                    KeyError(f"Note model '{holder.part_id}' to Csv config error. "
                             f"It expected {[field.name for field in model.fields]} but was missing: {missing}")
                )

            # Find mappings which do not exist on any note models
            if extra_fields:
                extra_fields = model.check_field_extra(extra_fields)

        if extra_fields:
            errors.append(
                KeyError(f"Field(s) '{extra_fields} are defined as mappings, but match no Note Model's field"))

        if errors:
            raise Exception(errors)

    def csv_row_map_to_note_fields(self, row: dict) -> dict:
        relevant_row_data = self.get_relevant_data(row)

        for pf in self.personal_fields:  # Add in Personal Fields
            relevant_row_data.setdefault(pf.field_name, False)
        for column in self.columns:  # Rename from Csv Column to Note Type Field
            if column.value in relevant_row_data:
                relevant_row_data[column.field_name] = relevant_row_data.pop(column.value)

        return relevant_row_data

    def csv_headers_map_to_note_fields(self, row: list) -> list:
        return list(self.csv_row_map_to_note_fields({row_name: "" for row_name in row}).keys())

    def note_fields_map_to_csv_row(self, row):
        for column in self.columns:  # Rename from Note Type Field to Csv Column
            if column.field_name in row:
                row[column.value] = row.pop(column.field_name)

        for pf in self.personal_fields:  # Remove Personal Fields
            if pf.field_name in row:
                del row[pf.field_name]

        relevant_row_data = self.get_relevant_data(row)

        return relevant_row_data

    def get_relevant_data(self, row):
        relevant_columns = [field.value for field in self.columns]
        if not relevant_columns:
            return []

        cols = list(row.keys())

        # errors = [KeyError(f"Missing column {rel_col}") for rel_col in relevant_columns if rel_col not in cols]
        # if errors:
        #     raise Exception(errors)

        irrelevant_columns = [column for column in cols if column not in relevant_columns]
        if not irrelevant_columns:
            return row

        relevant_data = {key: row[key] for key in row if key not in irrelevant_columns}

        return relevant_data

    def field_values_in_note_model_order(self, note_model_name, fields_from_csv):
        return [fields_from_csv[f] if f in fields_from_csv else ""
                for f in self.note_models[note_model_name].part.field_names_lowercase
                ]



from dataclasses import dataclass, field
from typing import Optional, Union, List

from brain_brew.configuration.build_config.build_task import BuildPartTask
from brain_brew.configuration.part_holder import PartHolder
from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.representation.generic.html_file import HTMLFile
from brain_brew.representation.yaml.note_model import NoteModel
from brain_brew.representation.yaml.note_model_field import Field
from brain_brew.representation.yaml.note_model_template import Template


@dataclass
class NoteModelFromHTMLParts(BuildPartTask):
    @classmethod
    def task_name(cls) -> str:
        return r'note_model_from_html_parts'

    @classmethod
    def yamale_schema(cls) -> str:
        return f'''\
            part_id: str()
            model_id: str()
            css_file: str()
            fields: list(include('{Field.task_name()}'))
            templates: list(str())
            model_name: str(required=False)
            save_to_file: str(required=False)
        '''

    @classmethod
    def yamale_dependencies(cls) -> set:
        return {Field}

    @dataclass
    class Representation(RepresentationBase):
        part_id: str
        model_id: str
        css_file: str
        fields: List[dict]
        templates: List[dict]
        model_name: Optional[str] = field(default=None)
        save_to_file: Optional[str] = field(default=None)

    @classmethod
    def from_repr(cls, data: Union[Representation, dict]):
        rep: cls.Representation = data if isinstance(data, cls.Representation) else cls.Representation.from_dict(data)
        return cls(
            part_id=rep.part_id,
            model_id=rep.model_id,
            css=HTMLFile.create_or_get(rep.css_file).get_data(deep_copy=True),
            fields=list(map(Field.from_dict, rep.fields)),
            templates=list(holder.part for holder in map(PartHolder.from_file_manager, rep.templates)),
            model_name=rep.model_name or rep.part_id,
            save_to_file=rep.save_to_file
        )

    part_id: str
    model_id: str
    css: str
    fields: List[Field]
    templates: List[Template]
    model_name: str
    save_to_file: Optional[str]

    def execute(self):
        part = NoteModel(
            name=self.model_name,
            id=self.model_id,
            css=self.css,
            fields=self.fields,
            templates=self.templates,
            required_fields_per_template=[]
        )

        PartHolder.override_or_create(self.part_id, self.save_to_file, part)



# from unittest.mock import patch
#
# import pytest
#
# from brain_brew.representation.configuration.note_model_mapping import NoteModelMapping, FieldMapping
# from brain_brew.representation.generic.csv_file import CsvFile
# from tests.test_file_manager import get_new_file_manager
#
#
# @pytest.fixture(autouse=True)
# def run_around_tests():
#     get_new_file_manager()
#     yield
#
#
# @pytest.fixture()
# def nmm_test1_repr() -> NoteModelMapping.Representation:
#     return NoteModelMapping.Representation(
#         "Test Model",
#         {
#             "guid": "guid",
#             "tags": "tags",
#
#             "english": "word",
#             "danish": "otherword"
#         },
#         []
#     )
#
#
# @pytest.fixture()
# def nmm_test2_repr() -> NoteModelMapping.Representation:
#     return NoteModelMapping.Representation(
#         "Test Model",
#         {
#             "guid": "guid",
#             "tags": "tags",
#
#             "english": "word",
#             "danish": "otherword"
#         },
#         ["extra", "morph_focus"]
#     )
#
#
# @pytest.fixture()
# def nmm_test1(nmm_test1_repr) -> NoteModelMapping:
#     return NoteModelMapping.from_repr(nmm_test1_repr)
#
#
# @pytest.fixture()
# def nmm_test2(nmm_test2_repr) -> NoteModelMapping:
#     return NoteModelMapping.from_repr(nmm_test2_repr)
#
#
# class TestInit:
#     def test_runs(self):
#         nmm = NoteModelMapping.from_repr(NoteModelMapping.Representation("test", {}, []))
#         assert isinstance(nmm, NoteModelMapping)
#
#     @pytest.mark.parametrize("read_file_now, note_model, personal_fields, columns", [
#         (False, "note_model.json", ["x"], {"guid": "guid", "tags": "tags", "english": "word", "danish": "otherword"}),
#         (True, "model_model", [], {"guid": "guid", "tags": "tags"}),
#         (False, "note_model-json", ["x", "y", "z"], {"guid": "guid", "tags": "tags", "english": "word", "danish": "otherword"})
#     ])
#     def test_values(self, read_file_now, note_model, personal_fields, columns):
#         config = setup_nmm_config(note_model, columns, personal_fields)
#
#         def assert_dp_note_model(passed_file, read_now):
#             assert passed_file == note_model
#             assert read_now == read_file_now
#
#         with patch.object(DeckPartNoteModel, "create", side_effect=assert_dp_note_model) as mock_nm:
#
#             nmm = NoteModelMapping(config, read_now=read_file_now)
#
#             assert isinstance(nmm, NoteModelMapping)
#             assert len(nmm.columns) == len(columns)
#             assert len(nmm.personal_fields) == len(personal_fields)
#
#             assert mock_nm.call_count == 1
#
#
# class TestVerifyContents:
#     pass  # TODO
#
#
# class TestCsvRowNoteFieldConversion:
#     @staticmethod
#     def get_csv_row(): return {
#         "guid": "AAAA",
#         "tags": "nice card",
#
#         "english": "what",
#         "danish": "hvad"
#     }
#
#     @staticmethod
#     def get_note_field(): return{
#         "guid": "AAAA",
#         "tags": "nice card",
#
#         "word": "what",
#         "otherword": "hvad",
#         "extra": False,
#         "morph_focus": False
#     }
#
#     def test_csv_row_map_to_note_fields(self, nmm_test_with_personal_fields1):
#         assert nmm_test_with_personal_fields1.csv_row_map_to_note_fields(self.get_csv_row()) == self.get_note_field()
#
#     def test_note_fields_map_to_csv_row(self, nmm_test_with_personal_fields1):
#         assert nmm_test_with_personal_fields1.note_fields_map_to_csv_row(self.get_note_field()) == self.get_csv_row()
#
#
# class TestGetRelevantData:
#     def test_data_correct(self, nmm_test_with_personal_fields1: NoteModelMapping, csv_test1: CsvFile):
#         expected_relevant_columns = ["guid", "english", "danish", "tags"]
#         data = csv_test1.get_data()
#
#         for row in data:
#             relevant_data = nmm_test_with_personal_fields1.get_relevant_data(row)
#             assert len(relevant_data) == 4
#             assert list(relevant_data.keys()) == expected_relevant_columns
#
#     def test_data_missing_columns(self, nmm_test_with_personal_fields1: NoteModelMapping, csv_test1: CsvFile):
#         row_missing = {
#             "guid": "test",
#             "english": "test"
#         }
#         with pytest.raises(Exception) as e:
#             relevant_data = nmm_test_with_personal_fields1.get_relevant_data(row_missing)
#
#         errors = e.value.args[0]
#         assert len(errors) == 2
#         assert isinstance(errors[0], KeyError)
#         assert isinstance(errors[1], KeyError)
#         assert errors[0].args[0] == "Missing column tags"
#         assert errors[1].args[0] == "Missing column danish"
#
#
# class TestFieldMapping:
#     def test_init(self):
#         fm = FieldMapping(FieldMapping.FieldMappingType.COLUMN, "Csv_Row", "note_model_field")
#         assert isinstance(fm, FieldMapping)
#         assert (fm.field_name, fm.value) == ("csv_row", "note_model_field")



from dataclasses import dataclass
from typing import Dict, Type, List, Set

from brain_brew.build_tasks.crowd_anki.headers_from_crowdanki import HeadersFromCrowdAnki
from brain_brew.build_tasks.crowd_anki.media_group_from_crowd_anki import MediaGroupFromCrowdAnki
from brain_brew.build_tasks.crowd_anki.note_models_from_crowd_anki import NoteModelsFromCrowdAnki
from brain_brew.build_tasks.crowd_anki.notes_from_crowd_anki import NotesFromCrowdAnki
from brain_brew.build_tasks.csvs.notes_from_csvs import NotesFromCsvs
from brain_brew.build_tasks.deck_parts.from_yaml_part import NotesFromYamlPart, NoteModelsFromYamlPart, \
    MediaGroupFromYamlPart
from brain_brew.build_tasks.deck_parts.headers_from_yaml_part import HeadersFromYamlPart
from brain_brew.build_tasks.deck_parts.media_group_from_folder import MediaGroupFromFolder
from brain_brew.build_tasks.deck_parts.media_group_to_folder import MediaGroupsToFolder
from brain_brew.build_tasks.deck_parts.note_model_from_html_parts import NoteModelFromHTMLParts
from brain_brew.build_tasks.deck_parts.note_model_template_from_html_files import TemplateFromHTML
from brain_brew.configuration.build_config.build_task import BuildTask, BuildPartTask, TopLevelBuildTask
from brain_brew.configuration.build_config.recipe_builder import RecipeBuilder


@dataclass
class PartsBuilder(RecipeBuilder, TopLevelBuildTask):
    accepts_list_of_self: bool = False

    @classmethod
    def task_name(cls) -> str:
        return r'build_parts'

    @classmethod
    def task_regex(cls) -> str:
        return r'build_part[s]?'

    @classmethod
    def known_task_dict(cls) -> Dict[str, Type[BuildTask]]:
        return BuildPartTask.get_all_task_regex(cls.yamale_dependencies())

    @classmethod
    def from_repr(cls, data: List[dict]):
        if not isinstance(data, list):
            raise TypeError(f"PartsBuilder needs a list")
        return cls.from_list(data)

    def encode(self) -> dict:
        pass

    @classmethod
    def from_yaml_file(cls, filename: str):
        pass

    @classmethod
    def yamale_schema(cls) -> str:
        return cls.build_yamale_root_node(cls.yamale_dependencies())

    @classmethod
    def yamale_dependencies(cls) -> Set[Type[BuildPartTask]]:
        return {
            NotesFromCsvs,
            NotesFromYamlPart, HeadersFromYamlPart, NoteModelsFromYamlPart, MediaGroupFromYamlPart,
            MediaGroupFromFolder, MediaGroupsToFolder,
            NoteModelFromHTMLParts, TemplateFromHTML,
            HeadersFromCrowdAnki, MediaGroupFromCrowdAnki, NoteModelsFromCrowdAnki, NotesFromCrowdAnki
        }



import logging
from abc import ABCMeta
from dataclasses import dataclass
from typing import List, Optional, Dict, Set

from brain_brew.configuration.global_config import GlobalConfig
from brain_brew.interfaces.media_container import MediaContainer
from brain_brew.representation.yaml.yaml_object import YamlObject
from brain_brew.utils import find_media_in_field

FIELDS = 'fields'
GUID = 'guid'
TAGS = 'tags'
NOTE_MODEL = 'note_model'
FLAGS = "flags"
NOTES = "notes"
NOTE_GROUPINGS = "note_groupings"
MEDIA_REFERENCES = "media_references"


@dataclass
class GroupableNoteData(YamlObject, MediaContainer, metaclass=ABCMeta):
    note_model: Optional[str]
    tags: Optional[List[str]]

    def encode_groupable(self, data_dict):
        if self.note_model is not None:
            data_dict.setdefault(NOTE_MODEL, self.note_model)
        if self.tags is not None and self.tags != []:
            data_dict.setdefault(TAGS, self.tags)
        return data_dict


@dataclass
class Note(GroupableNoteData):
    @classmethod
    def from_yaml_file(cls, filename: str) -> 'Note':
        return cls.from_dict(cls.read_to_dict(filename))

    fields: List[str]
    guid: str
    flags: int
    # media_references: Optional[Set[str]]

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            fields=data.get(FIELDS),
            guid=data.get(GUID),
            note_model=data.get(NOTE_MODEL, None),
            tags=data.get(TAGS, None),
            flags=data.get(FLAGS, 0)
        )

    def encode(self) -> dict:
        data_dict: Dict[str, any] = {FIELDS: self.fields, GUID: self.guid}
        if self.flags is not None and self.flags != 0:
            data_dict.setdefault(FLAGS, self.flags)
        super().encode_groupable(data_dict)
        return data_dict

    def get_all_media_references(self) -> Set[str]:
        return {entry for field in self.fields for entry in find_media_in_field(field)}


@dataclass
class NoteGrouping(GroupableNoteData):
    notes: List[Note]

    @classmethod
    def from_yaml_file(cls, filename: str) -> 'NoteGrouping':
        return cls.from_dict(cls.read_to_dict(filename))

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            notes=list(map(Note.from_dict, data.get(NOTES))),
            note_model=data.get(NOTE_MODEL, None),
            tags=data.get(TAGS, None)
        )

    def encode(self) -> dict:
        data_dict = {}
        super().encode_groupable(data_dict)
        data_dict.setdefault(NOTES, [note.encode() for note in self.notes])
        return data_dict

    # TODO: Extract Shared Tags and Note Models

    def verify_groupings(self):
        errors = []
        if self.note_model is not None:
            if any([note.note_model for note in self.notes]):
                errors.append(ValueError(f"NoteGrouping for 'note_model' {self.note_model} has notes with 'note_model'."
                                         f" Please remove one of these."))
        return errors

    def get_all_known_note_model_names(self) -> set:
        return {self.note_model} if self.note_model else {note.note_model for note in self.notes}

    def get_all_media_references(self) -> Set[str]:
        all_media = set()
        for note in self.notes:
            media = note.get_all_media_references()
            all_media = all_media.union(media)
        return all_media

    def get_sorted_notes(self, sort_by_keys, reverse_sort, case_insensitive_sort=None):
        if case_insensitive_sort is None:
            case_insensitive_sort = GlobalConfig.get_instance().sort_case_insensitive

        if sort_by_keys:
            def sort_method(i: Note):
                def get_sort_tuple(attr_or_field):
                    if attr_or_field in [GUID, FLAGS, NOTE_MODEL, TAGS]:
                        value = getattr(i, attr_or_field)
                    elif isinstance(attr_or_field, int) and attr_or_field < len(i.fields):
                        value = i.fields[attr_or_field]
                    else:
                        value = ""
                        logging.warning(f"No known sort value for {attr_or_field}")

                    if not isinstance(value, str):
                        return True, False
                    return (value == "", value.lower()) if case_insensitive_sort else (value == "", value)

                return tuple(get_sort_tuple(column) for column in sort_by_keys)

            return sorted(self.notes, key=sort_method, reverse=reverse_sort)
        elif reverse_sort:
            return list(reversed(self.notes))

        return self.notes
    
    def get_all_notes_copy(self, sort_by_keys, reverse_sort, case_insensitive_sort=None) -> List[Note]:
        def join_tags(n_tags):
            if self.tags is None and n_tags is None:
                return []
            elif self.tags is None:
                return n_tags
            elif n_tags is None:
                return self.tags
            else:
                return [*n_tags, *self.tags]

        return [Note(
                    note_model=self.note_model or n.note_model,
                    tags=join_tags(n.tags),
                    fields=n.fields,
                    guid=n.guid,
                    flags=n.flags
                    # media_references=n.media_references or n.get_media_references()
               ) for n in self.get_sorted_notes(sort_by_keys, reverse_sort, case_insensitive_sort)]


@dataclass
class Notes(YamlObject, MediaContainer):
    note_groupings: List[NoteGrouping]

    @classmethod
    def from_yaml_file(cls, filename: str) -> 'Notes':
        return cls.from_dict(cls.read_to_dict(filename))

    @classmethod
    def from_dict(cls, data: dict):
        return cls(note_groupings=list(map(NoteGrouping.from_dict, data.get(NOTE_GROUPINGS))))

    @classmethod
    def from_list_of_notes(cls, notes: List[Note]):
        return cls(note_groupings=[NoteGrouping(note_model=None, tags=None, notes=notes)])  # TODO: Check grouping here

    def encode(self) -> dict:
        data_dict = {NOTE_GROUPINGS: [note_grouping.encode() for note_grouping in self.note_groupings]}
        return data_dict

    def get_all_known_note_model_names(self):
        return {nms for group in self.note_groupings for nms in group.get_all_known_note_model_names()}

    def get_all_media_references(self) -> Set[str]:
        all_media = set()
        for note_group in self.note_groupings:
            media = note_group.get_all_media_references()
            all_media = all_media.union(media)
        return all_media

    def get_sorted_notes_copy(self, sort_by_keys, reverse_sort, case_insensitive_sort=None):
        return [note for group in self.note_groupings
                for note in group.get_all_notes_copy(sort_by_keys, reverse_sort, case_insensitive_sort)]



from abc import ABCMeta
from dataclasses import dataclass
from typing import Union

from brain_brew.configuration.build_config.build_task import BuildPartTask
from brain_brew.configuration.part_holder import PartHolder
from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.representation.yaml.media_group import MediaGroup
from brain_brew.representation.yaml.note_model import NoteModel
from brain_brew.representation.yaml.notes import Notes
from brain_brew.representation.yaml.yaml_object import YamlObject


@dataclass
class FromYamlPartBase(BuildPartTask, metaclass=ABCMeta):
    part_type = None

    @classmethod
    def yamale_schema(cls) -> str:
        return f'''\
            part_id: str()
            file: str()
        '''

    @dataclass
    class Representation(RepresentationBase):
        part_id: str
        file: str

    @classmethod
    def from_repr(cls, data: Union[Representation, dict]):
        rep: cls.Representation = data if isinstance(data, cls.Representation) else cls.Representation.from_dict(data)

        return cls(
            part=PartHolder.override_or_create(
                part_id=rep.part_id, save_to_file=None, part=cls.part_type.from_yaml_file(rep.file))
        )

    def execute(self):
        pass

    part: YamlObject


@dataclass
class NotesFromYamlPart(FromYamlPartBase):
    @classmethod
    def task_name(cls) -> str:
        return r'notes_from_yaml_part'

    part_type = Notes


@dataclass
class NoteModelsFromYamlPart(FromYamlPartBase, BuildPartTask):
    @classmethod
    def task_name(cls) -> str:
        return r'note_models_from_yaml_part'

    @classmethod
    def task_regex(cls) -> str:
        return r'note_model[s]?_from_yaml_part'

    part_type = NoteModel


@dataclass
class MediaGroupFromYamlPart(FromYamlPartBase, BuildPartTask):
    @classmethod
    def task_name(cls) -> str:
        return r'media_group_from_yaml_part'

    part_type = MediaGroup



import pytest

from brain_brew.representation.json.json_file import JsonFile
from brain_brew.representation.yaml.note_model import NoteModel
from brain_brew.representation.yaml.note_model_field import Field
from brain_brew.representation.yaml.note_model_template import Template
from brain_brew.representation.yaml.yaml_object import YamlObject
from tests.test_files import TestFiles


# CrowdAnki Files --------------------------------------------------------------------------


@pytest.fixture
def ca_nm_data_word():
    return JsonFile.read_file(TestFiles.CrowdAnkiNoteModels.LL_WORD_COMPLETE)


@pytest.fixture
def ca_nm_word(ca_nm_data_word) -> NoteModel:
    return NoteModel.from_crowdanki(ca_nm_data_word)


@pytest.fixture
def ca_nm_data_word_required_only():
    return JsonFile.read_file(TestFiles.CrowdAnkiNoteModels.LL_WORD_COMPLETE_ONLY_REQUIRED)


@pytest.fixture
def ca_nm_word_required_only(ca_nm_data_word_required_only) -> NoteModel:
    return NoteModel.from_crowdanki(ca_nm_data_word_required_only)


@pytest.fixture
def ca_nm_data_word_no_defaults():
    return JsonFile.read_file(TestFiles.CrowdAnkiNoteModels.LL_WORD_COMPLETE_NO_DEFAULTS)


@pytest.fixture
def ca_nm_word_no_defaults(ca_nm_data_word_no_defaults) -> NoteModel:
    return NoteModel.from_crowdanki(ca_nm_data_word_no_defaults)


# Yaml Files --------------------------------------------------------------------------
@pytest.fixture
def nm_data_word_required_only():
    return YamlObject.read_to_dict(TestFiles.NoteModels.LL_WORD_ONLY_REQUIRED)


@pytest.fixture
def nm_data_word_no_defaults():
    return YamlObject.read_to_dict(TestFiles.NoteModels.LL_WORD_NO_DEFAULTS)


class TestCrowdAnkiNoteModel:
    class TestConstructor:
        def test_normal(self, ca_nm_word):
            model = ca_nm_word
            assert isinstance(model, NoteModel)

            assert model.name == "LL Word"
            assert isinstance(model.fields, list)
            assert len(model.fields) == 7
            assert all([isinstance(field, Field) for field in model.fields])

            assert isinstance(model.templates, list)
            assert len(model.templates) == 7
            assert all([isinstance(template, Template) for template in model.templates])

        def test_only_required(self, ca_nm_word_required_only):
            model = ca_nm_word_required_only
            assert isinstance(model, NoteModel)

        def test_manual_construction(self):
            model = NoteModel(
                "name",
                "23094149+8124+91284+12984",
                "css is garbage",
                [],
                [Field(
                    "field1"
                )],
                [Template(
                    "template1",
                    "{{Question}}",
                    "{{Answer}}"
                )]
            )

            assert isinstance(model, NoteModel)

    class TestEncodeAsCrowdAnki:
        def test_normal(self, ca_nm_word, ca_nm_data_word):
            model = ca_nm_word

            encoded = model.encode_as_crowdanki()

            assert encoded == ca_nm_data_word

        def test_only_required_uses_defaults(self, ca_nm_word_required_only,
                                             ca_nm_data_word, ca_nm_data_word_required_only):
            model = ca_nm_word_required_only

            encoded = model.encode_as_crowdanki()

            assert encoded != ca_nm_data_word_required_only
            assert encoded == ca_nm_data_word

    class TestEncodeAsDeckPart:
        def test_normal(self, ca_nm_word, ca_nm_data_word, ca_nm_data_word_required_only, nm_data_word_required_only):
            model = ca_nm_word

            encoded = model.encode()

            assert encoded != ca_nm_data_word
            assert encoded != ca_nm_data_word_required_only
            assert encoded == nm_data_word_required_only

        def test_only_required_uses_defaults(self, ca_nm_word_no_defaults, ca_nm_data_word_no_defaults, nm_data_word_no_defaults):
            model = ca_nm_word_no_defaults

            encoded = model.encode()

            # debug_write_part_to_file(model, TestFiles.NoteModels.LL_WORD_NO_DEFAULTS)

            assert encoded != ca_nm_data_word_no_defaults
            assert encoded == nm_data_word_no_defaults



from dataclasses import dataclass, field
from typing import Union, Optional

from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.representation.yaml.yaml_object import YamlObject


@dataclass
class GlobalConfig(YamlObject):
    __instance = None

    def encode(self) -> dict:
        pass

    @dataclass
    class Representation(RepresentationBase):
        sort_case_insensitive: Optional[bool] = field(default=False)
        join_values_with: Optional[str] = field(default=" ")

    sort_case_insensitive: bool
    join_values_with: str

    @classmethod
    def from_repr(cls, data: Union[Representation, dict]):
        rep: cls.Representation = data if isinstance(data, cls.Representation) else cls.Representation.from_dict(data)
        return cls(
            sort_case_insensitive=rep.sort_case_insensitive,
            join_values_with=rep.join_values_with
        )

    def __post_init__(self):
        if GlobalConfig.__instance is None:
            GlobalConfig.__instance = self
        else:
            raise Exception("Multiple GlobalConfigs created")

    @classmethod
    def from_yaml_file(cls, filename: str = "brain_brew_config.yaml") -> 'GlobalConfig':
        return cls.from_repr(cls.read_to_dict(filename))

    @classmethod
    def get_instance(cls) -> 'GlobalConfig':
        return cls.__instance

    @classmethod
    def clear_instance(cls):
        if cls.__instance:
            cls.__instance = None



from typing import List

from brain_brew.configuration.part_holder import PartHolder
from brain_brew.interfaces.media_container import MediaContainer
from brain_brew.representation.yaml.media_group import MediaGroup


def create_media_group_from_location(
        part_id: str,
        save_to_file: str,
        media_group: MediaGroup,
        groups_to_blacklist: List[MediaContainer],
        groups_to_whitelist: List[MediaContainer]
) -> MediaGroup:
    if groups_to_whitelist:
        white = list(set.union(*[container.get_all_media_references() for container in groups_to_whitelist]))
        media_group.filter_by_filenames(white, should_match=True)

    if groups_to_blacklist:
        black = list(set.union(*[container.get_all_media_references() for container in groups_to_blacklist]))
        media_group.filter_by_filenames(black, should_match=False)

    holder = PartHolder.override_or_create(part_id, save_to_file, media_group)
    return holder.part



from dataclasses import dataclass, field
from typing import List, Union, Optional

from brain_brew.configuration.build_config.build_task import BuildPartTask
from brain_brew.configuration.part_holder import PartHolder
from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.representation.yaml.media_group import MediaGroup
from brain_brew.transformers.media_group_save_to_location import save_media_groups_to_location


@dataclass
class MediaGroupsToFolder(BuildPartTask):
    @classmethod
    def task_name(cls) -> str:
        return r'save_media_groups_to_folder'

    @classmethod
    def task_regex(cls) -> str:
        return r"save_media_group[s]?_to_folder"

    @classmethod
    def yamale_schema(cls) -> str:
        return f'''\
            parts: list(str())
            folder: str()
            clear_folder: bool(required=False)
            recursive: bool(required=False)
        '''

    @dataclass
    class Representation(RepresentationBase):
        parts: List[str]
        folder: str
        clear_folder: Optional[bool] = field(default=None)
        recursive: Optional[bool] = field(default=None)

    @classmethod
    def from_repr(cls, data: Union[Representation, dict]):
        rep: cls.Representation = data if isinstance(data, cls.Representation) else cls.Representation.from_dict(data)
        return cls(
            parts=list(holder.part for holder in map(PartHolder.from_file_manager, rep.parts)),
            folder=rep.folder,
            clear_folder=rep.clear_folder or False,
            recursive=rep.recursive or False
        )

    parts: List[MediaGroup]
    folder: str
    clear_folder: bool
    recursive: bool

    def execute(self):
        save_media_groups_to_location(self.parts, self.folder, self.clear_folder, self.recursive)



from dataclasses import dataclass, field
from typing import Optional, Union, Set

from brain_brew.configuration.anki_field import AnkiField
from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.representation.yaml.yaml_object import YamlObject
from brain_brew.utils import find_media_in_field

NAME = AnkiField("name")
ORDINAL = AnkiField("ord", "ordinal")
QUESTION_FORMAT = AnkiField("qfmt", "question_format")
ANSWER_FORMAT = AnkiField("afmt", "answer_format")
BROWSER_ANSWER_FORMAT = AnkiField("bafmt", "browser_answer_format", default_value="")
BROWSER_QUESTION_FORMAT = AnkiField("bqfmt", "browser_question_format", default_value="")
DECK_OVERRIDE_ID = AnkiField("did", "deck_override_id", default_value=None)


@dataclass
class Template(RepresentationBase, YamlObject):
    @classmethod
    def from_yaml_file(cls, filename: str) -> 'Template':
        return cls.from_dict(cls.read_to_dict(filename))

    @dataclass
    class CrowdAnki(RepresentationBase):
        name: str
        qfmt: str
        afmt: str
        bqfmt: str = field(default=BROWSER_QUESTION_FORMAT.default_value)
        bafmt: str = field(default=BROWSER_ANSWER_FORMAT.default_value)
        ord: int = field(default=None)
        did: Optional[int] = field(default=None)

    name: str
    question_format: str
    answer_format: str
    question_format_in_browser: str = field(default=BROWSER_QUESTION_FORMAT.default_value)
    answer_format_in_browser: str = field(default=BROWSER_ANSWER_FORMAT.default_value)
    deck_override_id: Optional[int] = field(default=DECK_OVERRIDE_ID.default_value)

    @classmethod
    def from_crowdanki(cls, data: Union[CrowdAnki, dict]):
        ca: cls.CrowdAnki = data if isinstance(data, cls.CrowdAnki) else cls.CrowdAnki.from_dict(data)
        return cls(
            name=ca.name, question_format=ca.qfmt, answer_format=ca.afmt,
            question_format_in_browser=ca.bqfmt, answer_format_in_browser=ca.bafmt, deck_override_id=ca.did
        )

    def get_all_media_references(self) -> Set[str]:
        all_media = set()\
            .union(find_media_in_field(self.question_format))\
            .union(find_media_in_field(self.answer_format))\
            .union(find_media_in_field(self.question_format_in_browser))\
            .union(find_media_in_field(self.answer_format_in_browser))
        return all_media

    def encode_as_crowdanki(self, ordinal: int) -> dict:
        data_dict = {
            ANSWER_FORMAT.anki_name: self.answer_format,
            BROWSER_ANSWER_FORMAT.anki_name: self.answer_format_in_browser,
            BROWSER_QUESTION_FORMAT.anki_name: self.question_format_in_browser,
            DECK_OVERRIDE_ID.anki_name: self.deck_override_id,
            NAME.anki_name: self.name,
            ORDINAL.anki_name: ordinal,
            QUESTION_FORMAT.anki_name: self.question_format,
        }

        return data_dict

    def encode(self) -> dict:
        data_dict = {
            NAME.name: self.name,
            QUESTION_FORMAT.name: self.question_format,
            ANSWER_FORMAT.name: self.answer_format
        }

        BROWSER_QUESTION_FORMAT.append_name_if_differs(data_dict, self.question_format_in_browser)
        BROWSER_ANSWER_FORMAT.append_name_if_differs(data_dict, self.answer_format_in_browser)
        DECK_OVERRIDE_ID.append_name_if_differs(data_dict, self.deck_override_id)

        return data_dict



# from unittest.mock import patch
#
# import pytest
#
# from brain_brew.constants.build_config_keys import BuildConfigKeys
# from brain_brew.build_tasks.source_crowd_anki import SourceCrowdAnki, CrowdAnkiKeys
# from brain_brew.representation.json.crowd_anki_export import CrowdAnkiExport
# from brain_brew.representation.json.part_header import DeckPartHeader
# from brain_brew.representation.yaml.note_model_repr import DeckPartNoteModel
# from brain_brew.representation.json.part_notes import DeckPartNotes
#
#
# def setup_ca_config(file, media, useless_note_keys, notes, headers):
#     return {
#         CrowdAnkiKeys.FILE.value: file,
#         CrowdAnkiKeys.MEDIA.value: media,
#         CrowdAnkiKeys.USELESS_NOTE_KEYS.value: useless_note_keys,
#         BuildConfigKeys.NOTES.value: notes,
#         BuildConfigKeys.HEADERS.value: headers
#     }
#
#
# class TestConstructor:
#     @pytest.mark.parametrize("file, media, useless_note_keys, notes, headers, read_file_now", [
#         ("test", False, {}, "test.json", "header.json", False),
#         ("export1", True, {}, "test.json", "header.json", False),
#         ("json.json", False, {}, "test.json", "", True),
#         ("", False, {"__type__": "Note", "data": None, "flags": 0}, "test.json", "header.json", False)
#     ])
#     def test_runs(self, file, media, useless_note_keys, notes, headers, read_file_now, global_config):
#         config = setup_ca_config(file, media, useless_note_keys, notes, headers)
#
#         def assert_dp_header(passed_file, read_now):
#             assert passed_file == headers
#             assert read_now == read_file_now
#
#         def assert_dp_notes(passed_file, read_now):
#             assert passed_file == notes
#             assert read_now == read_file_now
#
#         def assert_ca_export(passed_file, read_now):
#             assert passed_file == file
#             assert read_now == read_file_now
#
#         with patch.object(DeckPartHeader, "create", side_effect=assert_dp_header) as mock_header, \
#                 patch.object(DeckPartNotes, "create", side_effect=assert_dp_notes) as mock_notes, \
#                 patch.object(CrowdAnkiExport, "create", side_effect=assert_ca_export) as ca_export:
#
#             source = SourceCrowdAnki(config, read_now=read_file_now)
#
#             assert isinstance(source, SourceCrowdAnki)
#             assert source.should_handle_media == media
#             assert source.useless_note_keys == useless_note_keys
#
#             assert mock_header.call_count == 1
#             assert mock_notes.call_count == 1
#             assert ca_export.call_count == 1
#
#
# @pytest.fixture()
# def source_crowd_anki_test1(global_config) -> SourceCrowdAnki:
#     with patch.object(DeckPartHeader, "create", return_value=None) as mock_header, \
#          patch.object(DeckPartNotes, "create", return_value=None) as mock_notes, \
#          patch.object(CrowdAnkiExport, "create", return_value=None) as mock_ca_export:
#
#         source = SourceCrowdAnki(
#             setup_ca_config("", False, {"__type__": "Note", "data": None, "flags": 0}, "", "")
#         )
#
#         # source.notes = dp_notes_test1
#         # source.headers = dp_headers_test1
#         # source.crowd_anki_export = ca_export_test1
#
#     return source
#
#
# class TestSourceToDeckParts:
#     def test_runs(self, source_crowd_anki_test1: SourceCrowdAnki, ca_export_test1,
#                   temp_dp_note_model_file, temp_dp_headers_file, temp_dp_notes_file,
#                   dp_note_model_test1, dp_headers_test1, dp_notes_test1):
#
#         # CrowdAnki Export it will use to write to the DeckParts
#         source_crowd_anki_test1.crowd_anki_export = ca_export_test1
#
#         # DeckParts to be written to (+ the NoteModel below)
#         source_crowd_anki_test1.headers = temp_dp_headers_file
#         source_crowd_anki_test1.notes = temp_dp_notes_file
#
#         def assert_note_model(name, data_override):
#             assert data_override == dp_note_model_test1.get_data()
#             return dp_note_model_test1
#
#         with patch.object(DeckPartNoteModel, "create", side_effect=assert_note_model) as mock_nm:
#             source_crowd_anki_test1.source_to_parts()
#
#             assert source_crowd_anki_test1.headers.get_data() == dp_headers_test1.get_data()
#             assert source_crowd_anki_test1.notes.get_data() == dp_notes_test1.get_data()
#
#             assert mock_nm.call_count == 1
#
#
# class TestDeckPartsToSource:
#     def test_runs(self, source_crowd_anki_test1: SourceCrowdAnki, temp_ca_export_file,
#                   ca_export_test1, dp_notes_test1, dp_headers_test1):
#         source_crowd_anki_test1.crowd_anki_export = temp_ca_export_file  # File to write result to
#
#         # DeckParts it will use (+ dp_note_model_test1, but it reads that in as a file)
#         source_crowd_anki_test1.headers = dp_headers_test1
#         source_crowd_anki_test1.notes = dp_notes_test1
#
#         source_crowd_anki_test1.parts_to_source()  # Where the magic happens
#
#         assert temp_ca_export_file.get_data() == ca_export_test1.get_data()



import logging
from dataclasses import dataclass
from typing import Optional, TypeVar, Generic

T = TypeVar('T')


@dataclass
class PartHolder(Generic[T]):
    part_id: str
    save_to_file: Optional[str]
    part: T

    file_manager = None

    @classmethod
    def get_file_manager(cls):
        if not cls.file_manager:
            from brain_brew.configuration.file_manager import FileManager
            cls.file_manager = FileManager.get_instance()
        return cls.file_manager

    @classmethod
    def from_file_manager(cls, part_id: str) -> T:
        return cls.get_file_manager().get_part(part_id)

    @classmethod
    def override_or_create(cls, part_id: str, save_to_file: Optional[str], part: T):
        fm = cls.get_file_manager()

        dp = fm.get_part_if_exists(part_id)
        if dp is None:
            dp = fm.register_part(PartHolder(part_id, save_to_file, part))
        else:
            logging.warning(f"Overwriting existing Deck Part '{part_id}'")
            dp.part = part
            dp.save_to_file = save_to_file

        dp.write_to_file()

        return dp

    def write_to_file(self):
        if self.save_to_file is not None:
            self.part.dump_to_yaml(self.save_to_file)



from abc import ABC, abstractmethod


class YamlRepr(ABC):
    @classmethod
    @abstractmethod
    def task_name(cls) -> str:
        pass

    @classmethod
    @abstractmethod
    def yamale_schema(cls) -> str:
        pass

    @classmethod
    def yamale_dependencies(cls) -> set:
        return set()

    @classmethod
    @abstractmethod
    def from_repr(cls, data: dict):
        pass



from dataclasses import dataclass
from typing import Optional, Union, List


@dataclass
class SharedBaseNotes:
    @staticmethod
    def _get_sort_order(sort_order: Optional[Union[str, List[str]]]):
        if isinstance(sort_order, list):
            return sort_order
        elif isinstance(sort_order, str):
            return [sort_order]
        return []

    @staticmethod
    def _get_reverse_sort(reverse_sort: Optional[bool]):
        return reverse_sort or False

    # sort_order: Optional[List[str]]
    # reverse_sort: Optional[bool]



from dataclasses import dataclass, field
from typing import Optional

from brain_brew.configuration.build_config.build_task import BuildPartTask
from brain_brew.configuration.part_holder import PartHolder
from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.representation.generic.html_file import HTMLFile
from brain_brew.representation.yaml.note_model_template import Template

html_separator = '\n\n--\n\n'


@dataclass
class TemplateFromHTML(BuildPartTask):
    @classmethod
    def task_name(cls) -> str:
        return r'note_model_template_from_html'

    @classmethod
    def task_regex(cls) -> str:
        return r'note_model_template[s]?_from_html'

    @classmethod
    def yamale_schema(cls) -> str:
        return f"""\
            part_id: str()
            html_file: str()
            template_name: str(required=False)
            browser_html_file: str(required=False)
            deck_override_id: int(required=False)
            save_to_file: str(required=False)
        """

    @dataclass
    class Representation(RepresentationBase):
        part_id: str
        html_file: str
        template_name: Optional[str] = field(default=None)
        browser_html_file: Optional[str] = field(default=None)
        deck_override_id: Optional[int] = field(default=None)
        save_to_file: Optional[str] = field(default=None)

    @classmethod
    def from_repr(cls, data: dict):
        rep: cls.Representation = data if isinstance(data, cls.Representation) else cls.Representation.from_dict(data)
        return cls(
            part_id=rep.part_id,
            template_name=rep.template_name or rep.part_id,
            html_file=HTMLFile.create_or_get(rep.html_file),
            browser_html_file=HTMLFile.create_or_get(rep.browser_html_file) if rep.browser_html_file else None,
            deck_override_id=rep.deck_override_id,
            save_to_file=rep.save_to_file
        )

    part_id: str
    template_name: str
    html_file: HTMLFile
    browser_html_file: Optional[HTMLFile]
    deck_override_id: Optional[int]
    save_to_file: Optional[str]

    def execute(self):
        main_data = self.html_file.get_data(deep_copy=True)
        browser_data = self.browser_html_file.get_data(deep_copy=True) if self.browser_html_file else None

        if html_separator not in main_data:
            raise ValueError(f"Cannot find separator {html_separator} in html file '{self.html_file.file_location}'")

        front, back = tuple(main_data.split(html_separator, maxsplit=1))

        if browser_data:
            if html_separator not in browser_data:
                raise ValueError(f"Cannot find separator {html_separator} in html file '{self.browser_html_file.file_location}'")

            browser_front, browser_back = tuple(browser_data.split(html_separator, maxsplit=1))
        else:
            browser_front = browser_back = ""

        template = Template(
            name=self.template_name,
            question_format=front,
            answer_format=back,
            question_format_in_browser=browser_front,
            answer_format_in_browser=browser_back,
            deck_override_id=self.deck_override_id
        )

        PartHolder.override_or_create(self.part_id, self.save_to_file, template)



from abc import ABCMeta, abstractmethod
from typing import Dict, Type, Set

from brain_brew.interfaces.yamale_verifyable import YamlRepr


class BuildTask(YamlRepr, object, metaclass=ABCMeta):
    execute_immediately: bool = False
    accepts_list_of_self: bool = True

    @abstractmethod
    def execute(self):
        pass

    @classmethod
    def task_regex(cls) -> str:
        return cls.task_name()

    @classmethod
    def get_all_task_regex(cls, subclasses: Set[Type['BuildTask']]) -> Dict[str, Type['BuildTask']]:
        task_regex_matches: Dict[str, Type[BuildTask]] = {}

        for sc in subclasses:
            if sc.task_regex in task_regex_matches:
                raise KeyError(f"Multiple instances of task regex '{sc.task_regex}'")
            elif sc.task_regex == "" or sc.task_regex is None:
                raise KeyError(f"Unknown task regex in {sc.__name__}")

            task_regex_matches.setdefault(sc.task_regex(), sc)

        # logging.debug(f"Known build tasks: {known_build_tasks}")
        return task_regex_matches


class TopLevelBuildTask(BuildTask, metaclass=ABCMeta):
    pass


class BuildPartTask(BuildTask, metaclass=ABCMeta):
    execute_immediately: bool = True



from dataclasses import dataclass, field
from typing import List, Union

from brain_brew.configuration.anki_field import AnkiField
from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.interfaces.yamale_verifyable import YamlRepr

NAME = AnkiField("name")
ORDINAL = AnkiField("ord", "ordinal")
FONT = AnkiField("font", default_value="Liberation Sans")
MEDIA = AnkiField("media", default_value=[])
IS_RIGHT_TO_LEFT = AnkiField("rtl", "is_right_to_left", default_value=False)
FONT_SIZE = AnkiField("size", "font_size", default_value=20)
IS_STICKY = AnkiField("sticky", "is_sticky", default_value=False)


@dataclass
class Field(RepresentationBase, YamlRepr):
    @classmethod
    def task_name(cls) -> str:
        return r"note_model_field"

    @classmethod
    def yamale_schema(cls) -> str:
        return f"""\
            name: str()
            font: str(required=False)
            font_size: int(required=False)
            is_sticky: bool(required=False)
            is_right_to_left: bool(required=False)
        """

    @classmethod
    def from_repr(cls, data: dict):
        pass

    @dataclass
    class CrowdAnki(RepresentationBase):
        name: str
        ord: int = field(default=None)
        font: str = field(default=FONT.default_value)
        media: List[str] = field(default_factory=lambda: MEDIA.default_value)
        rtl: bool = field(default=IS_RIGHT_TO_LEFT.default_value)
        size: int = field(default=FONT_SIZE.default_value)
        sticky: bool = field(default=IS_STICKY.default_value)

    name: str
    font: str = field(default=FONT.default_value)
    is_right_to_left: bool = field(default=IS_RIGHT_TO_LEFT.default_value)
    font_size: int = field(default=FONT_SIZE.default_value)
    is_sticky: bool = field(default=IS_STICKY.default_value)
    media: List[str] = field(default_factory=lambda: MEDIA.default_value)  # Unused in Anki

    @classmethod
    def from_crowd_anki(cls, data: Union[CrowdAnki, dict]):
        ca: cls.CrowdAnki = data if isinstance(data, cls.CrowdAnki) else cls.CrowdAnki.from_dict(data)
        return cls(
            name=ca.name, font=ca.font, media=ca.media,
            is_right_to_left=ca.rtl, font_size=ca.size, is_sticky=ca.sticky
        )

    def encode_as_crowdanki(self, ordinal: int) -> dict:
        data_dict = {
            FONT.anki_name: self.font,
            MEDIA.anki_name: self.media,
            NAME.anki_name: self.name,
            ORDINAL.anki_name: ordinal,
            IS_RIGHT_TO_LEFT.anki_name: self.is_right_to_left,
            FONT_SIZE.anki_name: self.font_size,
            IS_STICKY.anki_name: self.is_sticky
        }

        return data_dict

    def encode_as_part(self) -> dict:
        data_dict = {
            NAME.name: self.name
        }

        FONT.append_name_if_differs(data_dict, self.font)
        MEDIA.append_name_if_differs(data_dict, self.media)
        IS_RIGHT_TO_LEFT.append_name_if_differs(data_dict, self.is_right_to_left)
        FONT_SIZE.append_name_if_differs(data_dict, self.font_size)
        IS_STICKY.append_name_if_differs(data_dict, self.is_sticky)

        return data_dict



import logging
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Union

from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.interfaces.yamale_verifyable import YamlRepr
from brain_brew.representation.generic.csv_file import CsvFile, CsvKeys
from brain_brew.utils import single_item_to_list, generate_anki_guid

FILE = "csv_file"
NOTE_MODEL = "note_model"
SORT_BY_COLUMNS = "sort_by_columns"
REVERSE_SORT = "reverse_sort"
DERIVATIVES = "derivatives"


@dataclass
class FileMappingDerivative(YamlRepr):
    @classmethod
    def task_name(cls) -> str:
        return r'file_mapping'

    @classmethod
    def yamale_schema(cls) -> str:
        return f'''\
            file: str()
            note_model: str(required=False)
            sort_by_columns: list(str(), required=False)
            reverse_sort: bool(required=False)
            derivatives: list(include('{cls.task_name()}'), required=False)
        '''

    @dataclass(init=False)
    class Representation(RepresentationBase):
        file: str
        note_model: Optional[str]
        sort_by_columns: Optional[Union[list, str]]
        reverse_sort: Optional[bool]
        derivatives: Optional[List['FileMappingDerivative.Representation']]

        def __init__(self, file, note_model=None, sort_by_columns=None, reverse_sort=None, derivatives=None):
            self.file = file
            self.note_model = note_model
            self.sort_by_columns = sort_by_columns
            self.reverse_sort = reverse_sort
            self.derivatives = list(map(FileMappingDerivative.Representation.from_dict, derivatives)) \
                if derivatives is not None else []

    compiled_data: Dict[str, dict] = field(init=False)

    csv_file: CsvFile

    note_model: Optional[str]
    sort_by_columns: Optional[list]
    reverse_sort: Optional[bool]
    derivatives: Optional[List['FileMappingDerivative']]

    @classmethod
    def from_repr(cls, data: Union[Representation, dict]):
        rep: cls.Representation = data if isinstance(data, cls.Representation) else cls.Representation.from_dict(data)
        return cls(
            csv_file=CsvFile.create_or_get(rep.file),
            note_model=rep.note_model.strip() if rep.note_model else None,
            sort_by_columns=single_item_to_list(rep.sort_by_columns),
            reverse_sort=rep.reverse_sort or False,
            derivatives=list(map(cls.from_repr, rep.derivatives)) if rep.derivatives is not None else []
        )

    def get_available_columns(self):
        return self.csv_file.column_headers + [col for der in self.derivatives for col in der.get_available_columns()]

    def get_used_note_model_names(self) -> List[str]:
        nm = [self.note_model] if self.note_model is not None else []
        return nm + [name for der in self.derivatives for name in der.get_used_note_model_names()]

    def _build_data_recursive(self) -> List[dict]:
        data_in_progress = self.csv_file.get_data(deep_copy=True)

        new_columns_seen_so_far = self.csv_file.column_headers.copy()
        for der in self.derivatives:
            der_cols = der.get_available_columns()
            overlapping_cols = [col for col in der_cols if col in self.csv_file.column_headers]
            der_cols = [col for col in der_cols if col not in overlapping_cols]

            if not overlapping_cols:
                raise KeyError("No column overlap for derivative")

            column_repeat_errors = [KeyError(f"Derivative column {c} in multiple derivative lines")
                                    for c in der_cols if c in new_columns_seen_so_far]
            if column_repeat_errors:
                raise Exception(column_repeat_errors)
            new_columns_seen_so_far += der_cols

            der_match_errors = []
            for der_row in der._build_data_recursive():
                # Find matching row to pair data with
                found_match = False
                for row in data_in_progress:
                    if all([der_row[c] == row[c] for c in overlapping_cols]):
                        for der_col in der_cols:
                            row[der_col] = der_row[der_col]
                        found_match = True
                        # Set Note Model to matching Derivative Note Model
                        if der.note_model is not None:
                            row.setdefault(NOTE_MODEL, der.note_model)
                        break
                if not found_match:
                    der_match_errors.append(ValueError(f"Cannot match derivative row {der_row} to parent"))

            if der_match_errors:
                raise Exception(der_match_errors)

        return data_in_progress

    def write_to_csv(self, data_to_set):
        self.csv_file.set_data_from_superset(data_to_set)
        self.csv_file.sort_data(self.sort_by_columns, self.reverse_sort)
        self.csv_file.write_file()

        for der in self.derivatives:
            der.write_to_csv(data_to_set)


@dataclass
class FileMapping(FileMappingDerivative):
    note_model: str  # Override Optional on Parent

    data_set_has_changed: bool = field(init=False, default=False)

    def compile_data(self):
        self.compiled_data = {}
        self.data_set_has_changed = False

        data_in_progress = self._build_data_recursive()

        # Set Note Model if not already set
        if self.note_model is not None:
            for row in data_in_progress:
                row.setdefault(NOTE_MODEL, self.note_model)

        # Fill in Guid if no Guid
        guids_generated = 0
        for row in data_in_progress:
            guid = row[CsvKeys.GUID.value]
            if not guid:
                guid = row[CsvKeys.GUID.value] = generate_anki_guid()
                guids_generated += 1
            self.compiled_data.setdefault(guid, {key.lower(): row[key] for key in row})

        if guids_generated > 0:
            self.data_set_has_changed = True
            logging.info(f"Generated {guids_generated} guids in {self.csv_file.file_location}")

    def set_relevant_data(self, data_set: Dict[str, dict]):
        unchanged, changed, added = 0, 0, 0
        for guid in data_set:
            if guid in self.compiled_data.keys():
                changed_row = False
                for key in data_set[guid]:
                    if key in self.compiled_data[guid] and self.compiled_data[guid][key] != data_set[guid][key]:
                        self.compiled_data[guid][key] = data_set[guid][key]
                        changed_row = True
                if changed_row:
                    changed += 1
                else:
                    unchanged += 1
            else:
                added += 1
                self.compiled_data.setdefault(guid, data_set[guid])

        if changed > 0 or added > 0:
            self.data_set_has_changed = True

        logging.info(f"Set {self.csv_file.file_location} data; changed {changed}, "
                     f"added {added}, while {unchanged} were identical")

    def write_file_on_close(self):
        self.write_to_csv(list(self.compiled_data.values()))
